# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import re

def build_dialog_node_index(workspace):
    dialog_nodes_by_id = {}
    for dialog_node in workspace['dialog_nodes']:
        dialog_nodes_by_id[dialog_node["dialog_node"]] = dialog_node     
    return dialog_nodes_by_id

class DialogNodeNotFound(KeyError):
    def __init___(self,skill, node):
        KeyError.__init__(self,"Skill or Node not found in Dialog {0}:{1}".format(skill, node))
        self.skill = skill
        self.node = node

class WA_Assistant_Skills:
    def __init__(self, name=""):
        self._assistant_name = name
        self._skills = {}
        
    def add_skill(self, skill_id, workspace):
        """Add a skill and create an index to fast lookup nodes by id

        The index field named "dialog_nodes_by_id" and it will be added to the rest of the WA fields:
        name, intents, entities, language, metadata, description, dialog_nodes, workspace_id, counter_examples. learning_opt_out, status
        
        Parameters
        ----------
        skill_id : str
            The file location of the spreadsheet
        workspace : dictionary
            the workspace object as received by WA v1 API
        """

        if skill_id in self._skills.keys():
            raise ValueError("Duplicate skill id: {} ".format(skill_id))
        workspace["dialog_nodes_by_id"] = build_dialog_node_index(workspace)
        self._skills[skill_id] = workspace

    def get_skill_by_id(self, skill_id):
        return self._skills[skill_id]
    
    def get_skill_name_by_id(self, skill_id, strict=False):
        if strict == True: 
            return self._skills[skill_id]["name"]
        if skill_id in self._skills:
            return self.get_skill_by_id(skill_id)["name"]
        else: 
            return "?"
    def get_name(self):
        return self._assistant_name

    def is_single_skill(self):
        return len(self._skills.keys()) == 1
    
    def get_single_skill(self):
        if not self.is_single_skill():
            raise ValueError("found more than one skill. This function supports only a single skill assistant")
        for value in self._skills.values():
            first_skill = value
            return first_skill

    def get_node_by_id(self, skill_id, node_id):
        skill = self.get_skill_by_id(skill_id)
        node = skill["dialog_nodes_by_id"][node_id]
        return node


    def find_first_skill_id_for_node(self, node_id):
        """
        Return the first skill_id that includes the node, otherwise None
        """
        for skill_id in self._skills:
            skill = self.get_skill_by_id(skill_id)
            if node_id in skill["dialog_nodes_by_id"]:
                return skill_id
        return None

    def get_node_by_id_any_skill(self, node_id):
        """
        find the first skill that includes the node_id definition and return the node in that skill.
        This function is useful for checking for nodes in multiple versiosn of skills, or when a user does not 
        want to match against the skill_id field, e.g. if he doesn't have access to all workspaces files
        """
        for skill_id in self._skills:
            skill = self.get_skill_by_id(skill_id)
            if node_id in skill["dialog_nodes_by_id"]:
                return skill["dialog_nodes_by_id"][node_id]
        #TBD return None or throw exception?
        raise DialogNodeNotFound("", node_id)

    def list_skills(self):
        return [self.get_skill_summary(skill_id) for skill_id in self._skills.keys()]
        #return list(self._skills.keys()) 
    
    def get_skill_summary(self, skill_id):
        skill = self.get_skill_by_id(skill_id)
        return { 
            'skill_id': skill_id, 
            'name': skill["name"], 
            'description': skill["description"] if "description" in skill else "", 
            'metadata': skill['metadata'], 
            'workspace_id': skill["workspace_id"],
            'num_of_intents': len(skill["intents"]),
            'num_of_entities': len(skill["entities"]),
            'num_of_dialog_nodes': len(skill["dialog_nodes"]),
            }
    
    def re_search_in_dialog_nodes(self, regex_str, keys=["title", "conditions", "dialog_node"], in_skill=None):
        """Find regular expression case insensitive search in nodes across the skills
        """

        results=[]
        for skill_id in self._skills:
            if in_skill != None:
                if skill_id != in_skill:
                    continue 
            for node in self._skills[skill_id]["dialog_nodes"]:
                for key in keys:
                    if key in node.keys():
                        #print (node[key])
                        match = re.search(regex_str, node[key], re.IGNORECASE)
                        if match:
                            result = node.copy()
                            result["skill_id"] = skill_id
                            result["matched_in"] = key 
                            result["matched_location"] = str(match.regs[0])
                            results.append(result)
        
        df = pd.DataFrame(columns=["skill_id", "dialog_node", "type", "title", "conditions", "matched_in","matched_location"])
        df = df.append(pd.DataFrame(results))
        
        return df[["skill_id", "dialog_node", "type", "title", "conditions", "matched_in","matched_location"]]



# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pandas as pd
import numpy as np
import decimal
import datetime
import re
import time, sys
from .wa_assistant_skills import DialogNodeNotFound

def ipython_info():
    ip = False
    if 'ipykernel' in sys.modules:
        ip = 'notebook'
    elif 'IPython' in sys.modules:
        ip = 'terminal'
    return ip

if ipython_info() == 'notebook':
    from IPython.display import clear_output


# TODO: the code supose to find last n
def extract_node_from_nodes_visited(nodes_visited, nodes_dict):
    num_nodes=len(nodes_visited)
    # last node is selected by default
    node_visited = nodes_visited[-1]
    for i in range(num_nodes-1, -1, -1):
        current_name = nodes_dict[nodes_visited[i]]
        if type(current_name) == str:
            if not current_name.startswith("node_1"): 
                # if we found another node with a "better" name
                node_visited = nodes_visited[i]
                break
        
    return node_visited
    

def to_canonical_WA(df, nodes_dict):
    """
    transform WA log format into pathflow dataframe canonical data model
    returns a canonical dataframe with the following mandatory columns
    'log_id'
    'conversation_id'
    'node_visited'
    'response_timestamp'
    'branch_exited'
    'branch_exited_reason'
    'intent'
    the following fields are optional:
    'response_type'
    'request_text'
    'response_text'
    """
    print("to_canonical_WA is deprecated and will be removed in a future release. Use to_canonical_WA_v2 instead")
    if df.empty:
        return df
    else:
        df1 = pd.concat([df.drop(['request', 'response'], axis=1).reset_index(drop=True),
                     df['request'].apply(pd.Series).add_prefix('request_').reset_index(drop=True),
                     pd.DataFrame(df['response']
                                  .tolist()).add_prefix('response_')], axis=1)
        df1['request_input'] = pd.io.json.json_normalize(df['request'])['input.text']
        df2 = pd.concat([df1.drop(['response_context', 'response_output'], axis=1), df1['response_context'].apply(pd.Series).add_prefix('response_context_'), pd.DataFrame(df1['response_output'].tolist()).add_prefix('response_output_')], axis=1)
        df3 = pd.concat([df2.drop(['response_context_system'], axis=1), df2['response_context_system'].apply(pd.Series).add_prefix('response_context_system_')], axis=1)

        if 'response_output_generic' in df3:
            df4 = pd.concat([df3.drop(['response_output_generic'], axis=1), df3['response_output_generic'].apply(pd.Series).add_prefix('response_output_generic_')], axis=1)
        else:
            df4 = df3

        cols = ['response_context_conversation_id', 'log_id' ,'response_timestamp', 
                'request_input', 'response_output_text' ,'response_output_nodes_visited', 
                'response_context_system_branch_exited', 'response_context_system_branch_exited_reason','response_intents'
                #,'response_output_generic_response_type']
                ]
        df5 = df4[cols].copy(deep=True)
        df5.rename(columns={'response_output_nodes_visited': 'nodes_visited',
                        'response_context_conversation_id': 'conversation_id',
                        'request_input' : 'request_text',
                        'response_output_text': 'response_text',
                        'response_context_system_branch_exited' : 'branch_exited',
                        'response_context_system_branch_exited_reason' : 'branch_exited_reason',
                        'response_intents' : 'intent'
                        #response_output_generic_response_type' : 'type'
                       }, inplace=True)
        # in the regular to_canonical, we always take the last node regardless if it's standard
        # TODO: consider only standard nodes
        #df5['node_visited']=df5['nodes_visited'].apply(lambda x:x[-1])
        
        # This string searches from the end for node with meaningful name
        df5['node_visited']=df5['nodes_visited'].apply(lambda x: extract_node_from_nodes_visited(x,nodes_dict)) 
        df5['response_timestamp'] = pd.to_datetime(df5['response_timestamp'])
        # TODO: check what happens for multiple and zero intents and reference by name
        df5['intent'] = df5['intent'].apply(lambda x:x[0]['intent'] if x else x)
        df5['node_visited'] = df5['node_visited'].replace(nodes_dict)
        return df5

def get_title_or_condition_or_id(node):
    if 'title' in node:
        return node["title"]
    elif 'conditions' in node:
        return node["conditions"]
    else: 
        return node["dialog_node"]

def extract_best_label(node, include_type=True):
    """
    Best label is defined based on node type as follows:
    response_condition: condition[RC]
    standard: title or condition[SN]
    frame: title[FR]
    event_handler: condition[EH]
    slot: variable[SL]
    folder: title[FOL]
    """
    def format(node_str, node_type):
        if not include_type:
            return node_str
        else:
            return node_str + "[" + node_type + "]"
    
    def get_event_handler_condition_or_type(node):
        if 'conditions' in node:
            return node["conditions"]
        else:
            return "[" + node["event_name"] + "]"

    try:
        if node["type"] == "response_condition":
            return format(get_title_or_condition_or_id(node), "RC")
        elif node["type"] == "standard":
            return format(get_title_or_condition_or_id(node), "SN")
        elif node["type"] == "folder":
            return format(node["title"], "FOL")
        elif node["type"] == "frame":
            return format(node["title"], "FR")
        elif node["type"] == "event_handler":
            return format(get_event_handler_condition_or_type(node), "EH")
        elif node["type"] == "slot":
            return format(node["variable"], "SL")
        else:
            return "?"
    except Exception as e:
        print("node label could not be resolved for ", e)
        return "?" 

def node_visited_to_str(row, wa_skills, include_node_visited_to_str):
    """
    returns a string representation of nodes_visited
    """
    results = []
    for node_id in row['nodes_visited']:
        try:
            node = wa_skills.get_node_by_id(row['skill_id'], node_id)
            results.append(extract_best_label(node, include_node_visited_to_str))
        except:
            results.append("?")
    return results

def extract_turn_label(row, wa_skills, errors, mode="last_standard", match_from_any_skill=False):
    """
    returns a string representation the most significant information in the turn
    mode="last_standard" will seek for the last standard node, and try to extract a meaningful name from it
    """
    #initialize with last node_id
    result=row['nodes_visited'][-1]
    for node_id in reversed(row['nodes_visited']):
        try:
            if match_from_any_skill == False:
                node = wa_skills.get_node_by_id(row['skill_id'], node_id)
            else:
                node = wa_skills.get_node_by_id_any_skill(node_id)
            if node["type"] == "standard":
                result = get_title_or_condition_or_id(node)
                if "title" in node:
                    #if we found a standard node with title we're done.  
                    break
                elif "conditions" in node:
                    if node["conditions"] != 'true':
                        #if we found a condition, check that it isn't true  If not keep on seeking for better name
                        break
            if node["type"] == "frame" or node["type"] == "folder":
                if "title" in node:
                    #if we find a frame or folder with a title, then we're done
                    result = node["title"]
                    break
        except KeyError as e:
            errors["skills_not_found"].add(e.args[0])
            errors["nodes_not_found"].add(node_id)
            #print("turn label could not be resolved for ", e.args[0], ":", node_id)
            result = "?"
            break
    return result

def is_intent_triggered(row, wa_skills, errors, match_from_any_skill=False):
    result = ''
    intent = row["intent_1"]

    for node_id in row['nodes_visited']:
        try:
            if intent == "":
                continue
            # TBD if strict, try to search in skill, if not strict, try to first search the skill, if failed, any skill
            # search in all skills
            #  is single skill, we could search in all skills, via get_node_by_id_anyskill
            if match_from_any_skill == False:
                node = wa_skills.get_node_by_id(row['skill_id'], node_id)
            else:
                node = wa_skills.get_node_by_id_any_skill(node_id)
            if "conditions" in node:
                if intent in node["conditions"]:
                    result = intent
                    break
        except KeyError as e:
            errors["skills_not_found"].add(e.args[0])
            errors["nodes_not_found"].add(node_id)
            #print("is_intent_triggered could not be resolved", e, ":", node_id)
            break
    return result

def to_canonical_WA_v2(df, wa_skills=None, skill_id_field=None, include_nodes_visited_str_types=False):
    """
    Transform WA log format into pathflow dataframe canonical data model

    Parameters:
    df (DataFrame): WA logs
    wa_skills (WA_Assistant_Skills)
    skill_id_field (str): the name of the column in the logs files to use to match with the skill.  If None, then the match will be done against all skills
    include_nodes_visited_str_types: include node type information in the ndoes_visited_str column

    Returns:
    Dataframe with the following mandatory columns
    'conversation_id', 'log_id', 'response_timestamp', 'request_text',
    'response_text', 'nodes_visited', 'branch_exited',
    'branch_exited_reason', 'skill_id', 'intent_1', 'intent_1_confidence',
    'event_type', 'intent_1_dialog_triggered', 'skill_name',
    'nodes_visited_str', 'turn_label'
    """

    # if wa_skills = None, only basic extraction will be performed, no enrichment will be performed
    # for turn_label or nodes_visited_str

    # if wa_skills is provided, the skill_id_field should define which field to match against.
    #  point to the skill_id_fieldno enrichments will b
    
    # if skill_id_field = None, assume single skill mode, and use get_node_by_id_anyskill(node_id) or find_first_skill_id_for_node to match


    should_enrich = True
    # validation
    if df.empty:
        print("Warning: empty dataframe, no logs to transform")
        return df
    if wa_skills == None or wa_skills.__class__.__name__ != 'WA_Assistant_Skills':
        print("Warning: WA assistant skill object not provided. Only basic transformation will be performed")
        should_enrich = False
    #TODO: 
    #if should_enrich == True and skill_id_field not in df.columns:
    #    print("Error: logs do not contain WA assistant skill object not provided.

    # CODE BELOW IMPROVES PERFORMANCE
    
    #expand the request and response objects into additional columns
    df1 = pd.concat([df.drop(['request', 'response'], axis=1).reset_index(drop=True),
            pd.DataFrame(df['request'].tolist()).add_prefix('request_'),
            pd.DataFrame(df['response'].tolist()).add_prefix('response_')], axis=1)
    #parse request_input
    df1['request_input'] = pd.io.json.json_normalize(df['request'])['input.text']
    
    #expand the request_context and response_output into additional columns
    df2 = pd.concat([df1.drop(['response_context', 'response_output'], axis=1), 
        pd.DataFrame(df1['response_context'].tolist()).add_prefix('response_context_'), 
        pd.DataFrame(df1['response_output'].tolist()).add_prefix('response_output_')], axis=1)

    #expand system_context columns (include fields e.g. the output text)
    df3 = pd.concat([df2.drop(['response_context_system'], axis=1), 
        pd.DataFrame(df2['response_context_system'].tolist()).add_prefix('response_context_system_')], axis=1)

    #expand response output generic columns (include fields e.g. the output text)  
    if 'response_output_generic' in df3:
        df4 = pd.concat([df3.drop(['response_output_generic'], axis=1), 
        pd.DataFrame(df3['response_output_generic'].tolist()).add_prefix('response_output_generic_')], axis=1)
    else:
        df4 = df3
    #TODO: consider to extract response_output_action too

    if skill_id_field != None:
        cols = ['response_context_conversation_id', 'log_id' ,'response_timestamp', 
        'request_input', 'response_output_text' ,'response_output_nodes_visited', 
        'response_context_system_branch_exited', 'response_context_system_branch_exited_reason',
        'response_intents',skill_id_field]
        #'response_output_generic_response_type']
    else: 
        cols = ['response_context_conversation_id', 'log_id' ,'response_timestamp', 
        'request_input', 'response_output_text' ,'response_output_nodes_visited', 
        'response_context_system_branch_exited', 'response_context_system_branch_exited_reason',
        'response_intents',"workspace_id"]
        #'response_output_generic_response_type']

    #if in_place == True:
    #    df5 = df4[cols]
    #else:
    df5 = df4[cols].copy(deep=True)
    
    if skill_id_field != None:
        df5.rename(columns={'response_output_nodes_visited': 'nodes_visited',
                        'response_context_conversation_id': 'conversation_id',
                        'request_input' : 'request_text',
                        'response_output_text': 'response_text',
                        'response_context_system_branch_exited' : 'branch_exited',
                        'response_context_system_branch_exited_reason' : 'branch_exited_reason',
                        'response_intents' : 'intent',
                        skill_id_field : 'skill_id'
                        #response_output_generic_response_type' : 'type'
                        }, inplace=True)
    else:
        df5.rename(columns={'response_output_nodes_visited': 'nodes_visited',
                        'response_context_conversation_id': 'conversation_id',
                        'request_input' : 'request_text',
                        'response_output_text': 'response_text',
                        'response_context_system_branch_exited' : 'branch_exited',
                        'response_context_system_branch_exited_reason' : 'branch_exited_reason',
                        'response_intents' : 'intent',
                        "workspace_id" : 'skill_id'
                        #response_output_generic_response_type' : 'type'
                        }, inplace=True)
    
    df5['response_timestamp'] = pd.to_datetime(df5['response_timestamp'])
    # TODO: check what happens for multiple and zero intents and reference by name
    df5['intent_1'] = df5['intent'].apply(lambda x:x[0]['intent'] if x else '')
    df5['intent_1_confidence'] = df5['intent'].apply(lambda x:x[0]['confidence'] if x else '')
    df5['event_type'] = "REQUEST_REPONSE"
    #df5['skill_id'] = pd.DataFrame(df[skill_id_field].to_list())
    
    errors = {
        "skills_not_found": set(),
        "nodes_not_found": set()
    }

    if should_enrich:
        if skill_id_field == None:
            df5['intent_1_dialog_triggered'] = df5.apply(lambda row: is_intent_triggered(row, wa_skills, errors, match_from_any_skill=True), axis=1)
        else:
            df5['intent_1_dialog_triggered'] = df5.apply(lambda row: is_intent_triggered(row, wa_skills, errors, match_from_any_skill=False), axis=1)
        
        if skill_id_field == None:
            df5['skill_name'] = df5['skill_id'].apply(lambda x: wa_skills.get_skill_name_by_id(x, False))
        else:
            try: 
                df5['skill_name'] = df5['skill_id'].apply(lambda x: wa_skills.get_skill_by_id(x)["name"])
            except KeyError as e:
                print("The skill_id: {} appears in the logs, but there is no corresponding skill_id in WA_Assistant_Skills object\n".format(str(e)))
                print("You should either add the missing skill, or set skill_id_field=None to try to match globally by node_ids that are part of WA_Assistant_Skill regatdless of the skill_id")
                raise e
        df5['nodes_visited_str']=df5.apply(lambda row: node_visited_to_str(row, wa_skills, include_nodes_visited_str_types), axis=1)
        
        if skill_id_field == None:
            df5['turn_label']=df5.apply(lambda row: extract_turn_label(row, wa_skills, errors, match_from_any_skill=True), axis=1)
        else:
            df5['turn_label']=df5.apply(lambda row: extract_turn_label(row, wa_skills, errors, match_from_any_skill=False), axis=1)
    df5.drop(['intent'], axis=1, inplace=True)
    if len(errors["skills_not_found"]) >0 or len(errors["nodes_not_found"]) > 0:
        print("Warning, transformation completed, but some values in the logs were not found in the corresponding WA_Assistant_Skills object:")
        print("skills not found: ", errors["skills_not_found"] )
        print("nodes not found: ", errors["nodes_not_found"] )

    return df5

def progress_bar(progress):
    bar_length = 30
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
    if progress < 0:
        progress = 0
    if progress >= 1:
        progress = 1
    block = int(round(bar_length * progress))
    if ipython_info() == 'notebook':
        clear_output(wait = True)
    text = "Progress: [{0}] {1:.1f}%".format( "#" * block + "-" * (bar_length - block), progress * 100)
    print(text)

def _to_canonical(df, nodes_dict):
    """
    The function does the following:
    1. manipulate the values in nodes_visted to not include brackets [] and apostrophes '' resulting in comma seperated list of dialog_node values
    2. Iterates over the nodes_visited array
    """
    if df.empty:
        return df
    else:
        cols = ['document_id', 'response_timestamp', 'conversation_id',
       'request_input', 'response_text', 'nodes_visited', 'response_intent',
       'branch_exited','branch_exited_reason', 'digression'] #,'node_visited'
        df_results = df[cols].copy(deep=True)

        import ast
        #transform string representation of nodes_visited into a list of nodes (array).  Note we're not using json.loads since
        #the data contains single quoted strings, so not valid json
        df_results['nodes_visited']=df_results['nodes_visited'].apply(lambda x:ast.literal_eval(x))
        df_results['node_visited']=df_results['nodes_visited']   
        for idx, element in enumerate(df_results['node_visited']):
            # extract a list of nodes_name (node_title) and list of node_types
            # if the node is not found in the dictionary (either bug, or wrong json workspace) then use the dialog_node
            # nodes_dict is a map between node_id --> (node_name, node_type)
            nodes_name = list(map(lambda x: nodes_dict[x][0] if x in nodes_dict else x, element))
            nodes_type = list(map(lambda x: nodes_dict[x][1] if x in nodes_dict else x, element))
        
            # if we have ["Login Issues", "node_12_lhagsdkljhsdlkg"]
            # the indcice_standard would produce [0]
            indice_standard = [i for i, y in enumerate(nodes_type) if y == "standard"]
            # check how many value are there in nodes_visited
            # special case, if last standard node (index -2) is Handoff or More Help - TBD double check with Erica
            # if we have  ["More Help", "Login Issues"], then select More Help
            if len (nodes_name)>2 and (nodes_name[-2]=='Handoff' or nodes_name[-2]=='More Help'):
                df_results.loc[idx, 'node_visited'] = nodes_name[-2]
            elif len (indice_standard) > 0:  # take the last standard node
                df_results.loc[idx, 'node_visited'] = nodes_name[indice_standard[-1]]  ## logs will compare ids to the title
            else:
                df_results.loc[idx, 'node_visited'] = nodes_name[-1]
        df_results.rename(columns={
                        'document_id': 'log_id',
                        'response_intent' : 'intent',
                        'request_input' : 'request_text'
                       }, inplace=True)
        df_results.rename(columns={'document_id': 'log_id'}, inplace=True)
        return df_results


# import re
# def _to_canonical(df, nodes_dict):
#     """
#     The function does the following:
#     1. manipulate the values in nodes_visted to not include brackets [] and apostrophes '' resulting in comma seperated list of dialog_node values
#     2. Iterates over the nodes_visited array
#     """

#     if df.empty:
#         return df
#     else:
#         # remove special characters.but the space could be a bug
#         # however in this example, ["Login Issues", "node_12_lhagsdkljhsdlkg"]
#         # the algorithm would remove the space
#         # ["LoginIssues", "node_12_lhagsdkljhsdlkg"] which is an issue since
#         # nodes_name includes the key "Login Issues"
#         #sym_list = str.maketrans("", "", "'[]' ")
#         sym_list = str.maketrans("", "", "'[]")
#         df['node_visited']=[x.translate(sym_list) for x in df['nodes_visited'].tolist()]
#         pattern = re.compile(",")
#         df['node_visited']=df['node_visited'].apply(lambda x:pattern.split(x))

#         for idx, element in enumerate(df['node_visited']):
#             # extract a list of nodes_name (node_title) and list of node_types
#             # if the node is not found in the dictionary (either bug, or wrong json workspace) then use the dialog_node
#             nodes_name = list(map(lambda x: list(nodes_dict[x])[0].lstrip() if x in nodes_dict else x.lstrip(), element))
#             nodes_type = list(map(lambda x: list(nodes_dict[x])[1].lstrip() if x in nodes_dict else x.lstrip(), element))
#             # if we have ["Login Issues", "node_12_lhagsdkljhsdlkg"]
#             # the indcice_standard would produce [0]
#             indice_standard = [i for i, y in enumerate(nodes_type) if y == "standard"]
#             # check how many value are there in nodes_visited
#             # special case, if last standard node (index -2) is Handoff or More Help - TBD double check with Erica
#             # if we have  ["More Help", "Login Issues"], then select More Help
#             if len (nodes_name)>2 and (nodes_name[-2]=='Handoff' or nodes_name[-2]=='More Help'):
#                 df.loc[idx, 'node_visited'] = nodes_name[-2]
#             elif len (indice_standard) > 0:  # take the last standard node
#                 df.loc[idx, 'node_visited'] = nodes_name[indice_standard[-1]]  ## logs will compare ids to the title
#             else:
#                 df.loc[idx, 'node_visited'] = nodes_name[-1]

#      # TODO: consider uncomment
#      #   if ipython_info() == 'notebook':
#             progress_bar(idx / len(df['node_visited']))
#         cols = ['document_id', 'response_timestamp', 'conversation_id',
#        'request_input', 'response_text', 'nodes_visited', 'response_intent',
#        'branch_exited','branch_exited_reason', 'digression','node_visited']
#         df1 = df[cols].copy(deep=True)
#         # TODO: talk with Doug on intent field. Does it always exist?
#         df1.rename(columns={
#                         'document_id': 'log_id',
#                         'response_intent' : 'intent',
#                         'request_input' : 'request_text'
#                        }, inplace=True)
#         df1.rename(columns={'document_id': 'log_id'}, inplace=True)
#         return df1


def conv_id_dict (df_logs_formated):
    UniqueConv = df_logs_formated.conversation_id.unique()
    ConvDict = {elem : pd.DataFrame for elem in UniqueConv}
    for key in ConvDict.keys():
        ConvDict[key] = df_logs_formated[:][df_logs_formated.conversation_id == key]
    return ConvDict

# TODO: consider using a single implementation for all cases
def extract_dialog_node_name_WA(workspace_nodes):
    """
    Extract more friendly dialog node names from the Watson Assistant workspace
    return a nodes dictionary object.
    """

    nodes_dict = {}
    nodes_type = {}
    for idx,obj in workspace_nodes.iterrows():
        if (obj['type']=='standard') and not (obj['title'] is np.nan or obj['title'] != obj['title']):
            nodes_dict[obj['dialog_node']]=obj['title']
        elif ('conditions' in obj) and (obj['conditions'] != "true"):
            nodes_dict[obj['dialog_node']]=obj['conditions']
        else:
            nodes_dict[obj['dialog_node']]=obj['dialog_node']
    return nodes_dict


def _extract_dialog_node_name(dialog_nodes):
    """
    For each dialog_node (node_id) of type *standard*, check if *title exists*.
    If exists, use the title for the node_name. otherwise, use the dialog_node
    For all other cases, use the dialog_node

    dialog_node: (dialog_node_title, dialog_node_type)

    In the case of Login Issues,
    "title": "Login Issue",
    "dialog_node": "Login Issues",

    the record will be created as:
    "Login Issues": ("Login Issue", "standard")
    """
    nodes_dict = {}
    nodes_type = {}
    for obj in dialog_nodes:
        if (obj['type']=='standard') and ('title' in obj):
            if (obj['title'] is not None):
                nodes_dict[obj['dialog_node']] = (obj['title'],obj['type'])
            else:
                nodes_dict[obj['dialog_node']] = (obj['dialog_node'],obj['type'])
        else:
            nodes_dict[obj['dialog_node']] = (obj['dialog_node'],obj['type'])
    return nodes_dict



# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from IPython.display import HTML
import random

def hide_cell_toggle(for_next=False):
    this_cell = """$('div.cell.code_cell.rendered.selected')"""
    next_cell = this_cell + '.next()'

    toggle_text = 'Toggle show/hide'  # text shown on toggle link
    target_cell = this_cell  # target cell to control with toggle
    js_hide_current = ''  # bit of JS to permanently hide code in current cell (only when toggling next cell)

    if for_next:
        target_cell = next_cell
        toggle_text += ' next cell'
        js_hide_current = this_cell + '.find("div.input").hide();'

    js_f_name = 'code_toggle_{}'.format(str(random.randint(1,2**64)))

    html = """
        <script>
            function {f_name}() {{
                {cell_selector}.find('div.input').toggle();
            }}

            {js_hide_current}
        </script>

        <a href="javascript:{f_name}()">{toggle_text}</a>
    """.format(
        f_name=js_f_name,
        cell_selector=target_cell,
        js_hide_current=js_hide_current,
        toggle_text=toggle_text
    )

    return HTML(html)



# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

def _events_from_row(canonical_row):
    events_array=[]
    current_request = canonical_row["request_text"]
    current_response = canonical_row["response_text"]
    if current_request!="":
        event={}
        event["agent"]="user"
        event["type"]="request"
        event["timestamp"]=canonical_row["response_timestamp"]
        current_message={}
        current_message["text"]=current_request
        current_messages=[current_message]
        event["messages"]=current_messages
        if "highlight" in canonical_row:
            if canonical_row["highlight"] == True:
                event["highlight"] = True                   
        events_array.append(event)

    # create response object
    event={}
    event["agent"]="bot"
    event["type"]="response"
    event["timestamp"]=canonical_row["response_timestamp"]
    current_message={}
    current_message["text"]=current_response
    if "node_visited" in canonical_row:
        event["node_visited"] = canonical_row["node_visited"]        
    if "bot_response_buttons" in canonical_row:
        # temporary add buttons as 2nd 
        s = canonical_row["bot_response_buttons"]
#         s = s.replace(']','')
#         s = s.replace('[','')
#         s = s.replace("'",'')
        if s != '[]':
            current_message["buttons"] = s
    if "bot_response_action" in canonical_row:
        if canonical_row["bot_response_action"]=="Transfer":
            event["type"]="handoff"
        event["action"] = canonical_row["bot_response_action"]  
    if 'turn_label' in canonical_row:
        event["turn_label"] = canonical_row["turn_label"]  
    if 'nodes_visited_str' in canonical_row:
        event["nodes_visited_str"] = canonical_row["nodes_visited_str"]  
    if 'nodes_visited' in canonical_row:
        event["nodes_visited"] = canonical_row["nodes_visited"]  
    if 'skill_name' in canonical_row:
        event["skill_name"] = canonical_row["skill_name"]  
    
    current_messages=[current_message]
    event["messages"]=current_messages 
    events_array.append(event)

    return events_array        

def _add_insights_tags(existing_tags, canonical_row):
    if "insights_tags" in canonical_row:
        new_tags = canonical_row["insights_tags"]
        for new_tag in new_tags:
            if new_tag in existing_tags:
                existing_tags[new_tag]+=1
            else:
                existing_tags[new_tag]=1            

def to_transcript(df_canonical):
    dict_list=[]
    df_canonical=df_canonical.sort_values(by=["conversation_id", "response_timestamp"])
    conversations=df_canonical.groupby("conversation_id")
    for name, group in conversations:
        current_dict={}
        current_conversation={}
        current_conversation["id"]=name
        current_dict["conversation"]=current_conversation
         
        conversation_events=[]
        insight_tags = {} # map of tags to their counts
        for row_index, row in group.iterrows():
            conversation_events.extend(_events_from_row(row))
            _add_insights_tags(insight_tags, row)
        current_dict["events"] = conversation_events
        current_dict["insight_tags"] = insight_tags
        dict_list.append(current_dict)
        
    return dict_list


# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

def read_logs(service, workspace_id, num_logs, log_filter=None, sort="-request_timestamp"):
    try:
        log_list=[]
        current_cursor = None
        while num_logs > 0:
            response = service.list_logs(
                workspace_id=workspace_id,
                page_limit=500,
                cursor=current_cursor,
                sort=sort,
                filter=log_filter
            ).get_result()
            min_num = min(num_logs, len(response['logs']))
            log_list.extend(response['logs'][:min_num])
            #print("\r{} logs retrieved'.format(len(log_list)), end=''")
            print("\r{} logs retrieved".format(len(log_list)))
            num_logs = num_logs - min_num
            current_cursor = None
            # Check if there is another page of logs to be fetched
            if 'pagination' in response:
                # Get the url from which logs are to fetched
                if 'next_cursor' in response['pagination']:
                    current_cursor = response['pagination']['next_cursor']
                else:
                    break
    except WatsonApiException:
        print('You\'ve reached the rate limit of log api, refer to https://www.ibm.com/watson/developercloud/assist'
              'ant/api/v1/curl.html?curl#list-logs for additional information')
    except Exception as ex:
        print(ex)
    finally:
        log_df = pd.DataFrame(log_list)
        return log_df



# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The source code for this program is not published or otherwise divested of its trade secrets, irrespective of what has been depos-ited with the U.S. Copyright Office.
########################################################################
# Functions that process flows

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

########################################################################
def get_nodes_list(main_list):
    global descendant_list
    if type(main_list) == list:
        for i in main_list:
            if type(i) == str:
                descendant_list.append(i)
            else:
                get_nodes_list(i)
    elif type(main_list) == dict:
        for i in main_list:
            if type(main_list[i]) == str:
                descendant_list.append(i)
            else:
                descendant_list.append(i)
                get_nodes_list(main_list[i])

##########################################################################
def find_node_in_tree(node_name, main_list):
    global descendant_list
    for i in main_list:
        if i == node_name:
            descendant_list.append(i)
            if type(main_list) == dict:
                get_nodes_list(main_list[i])
            elif type(main_list) == list:
                get_nodes_list(i)
    else:
        if type(main_list) == list:
            for i in main_list:
                find_node_in_tree(node_name, i)
        elif type(main_list) == dict:
            for i in main_list:
                find_node_in_tree(node_name, main_list[i])

####################################################################
def get_descendants(nodes_tree, nodes_names, skill=None):
    
    global descendant_list
    descendant_list = []
    
    for node in nodes_names:
        if skill == "all":
            for i in nodes_tree:
                find_node_in_tree(node, nodes_tree[i])
        elif skill is not None:
            find_node_in_tree(node, nodes_tree[skill])
    return descendant_list

#############################################################################
def detect_flow_state(nodes, flows):

    if type(nodes) != list:
        return None

    processed_found=False
    for flow in flows:
        for node in nodes:
            if node in flow['completion_nodes']:
                return "completed"
            elif node in flow['nodes']:
                processed_found=True
        if processed_found:
            return "processed"
    
    return None

#############################################################################
def detect_flow(nodes, flows):
#give to it flow_defs['flows'] and node as list, for example ['node_xxxxxxxxx']
# return latest one
    if type(nodes) != list:
        return None
    
    for flow in flows:
        for node in nodes:
            if node in flow['nodes']:
                return flow['name']
    
    return None

###########################################################################################
# add internal flow content to initial definition
def enrich_flows_by_workspace(flow_defs, workspace, workspace_name="Some workspace"):

    workspace_edited={}
    workspace_edited[workspace_name]={}
    workspace_edited[workspace_name]["workspace_name"]=workspace["name"]
    workspace_edited[workspace_name]['indexed_dialog_nodes']={}

    for i in range(0,len(workspace["dialog_nodes"])):
        workspace_edited[workspace_name]['indexed_dialog_nodes'][workspace["dialog_nodes"][i]["dialog_node"]] =\
            workspace["dialog_nodes"][i]   
    
    tree_dict = {}
    for key in workspace_edited.keys():
        par = got_parent_nodes_for_skill(key, workspace_edited)  
        tree_dict[key] = create_nodes_tree(par, workspace_edited)
    
    for i in range (0, len(flow_defs["flows"])):        
        flow_defs["flows"][i]["nodes"] = get_descendants(tree_dict, flow_defs["flows"][i]["parent_nodes"], 'all')
    
    return flow_defs
    
#######################################################################################
# add two flow columns to canonical dataframe
def enrich_canonical_by_flows(df_all, flow_defs):
    df_all["flow"] = df_all.nodes_visited.map(lambda x: detect_flow(x, flow_defs["flows"]))
    df_all["flow_state"] = df_all.nodes_visited.map(lambda x: detect_flow_state(x, flow_defs["flows"]))
    
    return df_all

#######################################################################################
def got_parent_nodes_for_skill(skill,workspaces):
    nodes_parents = {}
    for i in workspaces[skill]['indexed_dialog_nodes']:
        if "parent" in workspaces[skill]['indexed_dialog_nodes'][i].keys():
            if workspaces[skill]['indexed_dialog_nodes'][i]['parent'] not in nodes_parents:
                nodes_parents[workspaces[skill]['indexed_dialog_nodes'][i]['parent']] = []
            nodes_parents[workspaces[skill]['indexed_dialog_nodes'][i]['parent']].append(workspaces[skill]['indexed_dialog_nodes'][i]['dialog_node'])
    return nodes_parents

def create_nodes_tree(parents_dict,workspaces):
    global delete_list
    delete_list = []
    new_dict = {}
    delete_list = []
    for i in parents_dict:
        values = parents_dict[i]
        #check if master
        new_values = check_if_master(values, parents_dict)
        new_dict[i] = new_values
    
    
    for z in set(delete_list):
        del new_dict[z]
        
    return new_dict

 
def check_if_master(values, parents_dict):
    new_values = []
    for i in values:
        if i in parents_dict:
            #it's master
            new_values.append({i : check_if_master(parents_dict[i], parents_dict)})
            global delete_list
            delete_list.append(i)
        else:
            #not a master
            new_values.append(i)
    return new_values
 

##########################################################################################################
 # compute flow statistics
def count_flows(df_logs, flow_defs): 
    
    df_logs = df_logs.sort_values(["conversation_id", "response_timestamp"]) 
    
    flow_outcome_summary={}
    flow_names=[]
    
    for i in range(0,len(flow_defs["flows"])):
        flow_names.append(flow_defs["flows"][i]["name"])
        flow_outcome_summary[flow_names[i]] = {}
        flow_outcome_summary[flow_names[i]]["completed"]=0
        flow_outcome_summary[flow_names[i]]["abandoned"]=0
        flow_outcome_summary[flow_names[i]]["rerouted"]=0
        #flow_outcome_summary[flow_names[i]]["escalated"]=0 
        
    num_flows=len(flow_names) 
    
    previous_conversation_id=""
    previous_flow_state=""
    previous_flow=""
    # state of each flow will be monitored
    current_conversation_flow_state = {}
    for i in range(0, num_flows):
        current_conversation_flow_state[flow_names[i]] = ""

    for i in range(0,len(df_logs)):
       
        current_conversation_id = df_logs.iloc[i,df_logs.columns.get_loc("conversation_id")]
        current_flow =  df_logs.iloc[i,df_logs.columns.get_loc("flow")]  
        current_flow_state =  df_logs.iloc[i,df_logs.columns.get_loc("flow_state")]  
        
              
        if current_conversation_id!=previous_conversation_id:
            # resetting states of all flows
            for j in range(0, num_flows):
                current_conversation_flow_state[flow_names[j]] = ""
        
            previous_conversation_id=current_conversation_id 
            new_session=True
        else:
            new_session=False
        
        if current_flow is not None:

            if current_flow_state=="processed":
                if current_conversation_flow_state[current_flow]=="" or new_session: # started now
                    current_conversation_flow_state[current_flow]="processed"
                    flow_outcome_summary[current_flow]["abandoned"]+=1  # abandonment is initial default
                elif current_conversation_flow_state[current_flow]=="rerouted": # return of digression
                    current_conversation_flow_state[current_flow]="processed"
                    flow_outcome_summary[current_flow]["abandoned"]+=1  # abandonment is initial default
                    flow_outcome_summary[current_flow]["rerouted"]-=1

            elif current_flow_state=="completed":        
                flow_outcome_summary[current_flow]["completed"]+=1
                if current_conversation_flow_state[current_flow]=="processed":
                    flow_outcome_summary[current_flow]["abandoned"]-=1
                elif current_conversation_flow_state[current_flow]=="rerouted":
                    flow_outcome_summary[current_flow]["rerouted"]-=1
                current_conversation_flow_state[current_flow]=""   
            
            else:
                print("Error! Wrong flow state.")
            
        # TODO: Check that dashboard code is fine
        if not new_session:
            if (previous_flow!=current_flow or current_flow is None) and previous_flow_state=="processed":
                flow_outcome_summary[previous_flow]["rerouted"]+=1
                current_conversation_flow_state[previous_flow]="rerouted"
                flow_outcome_summary[previous_flow]["abandoned"]-=1 
                                        
        previous_flow=current_flow      
        previous_flow_state=current_flow_state   
    
    # Temporarily we will not have escalated
             
    for key in flow_outcome_summary.keys():
        flow_outcome_summary[key]["overall"] = flow_outcome_summary[key]["completed"] +\
            flow_outcome_summary[key]["rerouted"] + flow_outcome_summary[key]["abandoned"]
            
    df=pd.DataFrame.from_dict(flow_outcome_summary, orient='index',
                       columns=['overall','completed', 'abandoned', 'rerouted'])

    df.index.name="flow name"

    df.reset_index(inplace=True)
        
    return df

################################################################################################
# flow chart for flow completion / abandonment / rerouting
def plot_flow_outcomes(df_flow_outcome_summary):
 
    flow_names =  df_flow_outcome_summary["flow name"].tolist()
    num_flows=len(flow_names)
    
    flow_outcome_summary = df_flow_outcome_summary.set_index('flow name').T.to_dict() 
    category_names=["Completed", "Rerouted", "Abandoned"]
    
    for key in flow_outcome_summary.keys():
        flow_outcome_summary[key]["overall"] = flow_outcome_summary[key]["completed"] +\
            flow_outcome_summary[key]["rerouted"] + flow_outcome_summary[key]["abandoned"]
    
    labels=[]
    for i in range(0, num_flows):
        labels.append(flow_names[i] + (" ("+str(flow_outcome_summary[flow_names[i]]["overall"])+")"))
        
    data = np.zeros([num_flows,3])    
    
    for i in range(0, num_flows):
        data[i,0] = (flow_outcome_summary[flow_names[i]]["completed"] / flow_outcome_summary[flow_names[i]]["overall"]) * 100.0
        data[i,1] = (flow_outcome_summary[flow_names[i]]["rerouted"] / flow_outcome_summary[flow_names[i]]["overall"]) * 100.0
        data[i,2] = (flow_outcome_summary[flow_names[i]]["abandoned"] / flow_outcome_summary[flow_names[i]]["overall"]) * 100.0
        
    data_cum = data.cumsum(axis=1)

    alpha = 0.7
    category_colors = [(0.102, 0.596, 0.314,alpha),(0.878, 0.878, 0.878, alpha),(0.843, 0.188, 0.152, alpha)]
    fig, ax = plt.subplots(figsize=(12, 2))
    ax.invert_yaxis()
    #ax.xaxis.set_visible(False)
    ax.set_xlim(0, np.sum(data, axis=1).max())
    
    for i, (colname, color) in enumerate(zip(category_names, category_colors)):
        widths = data[:, i]
        starts = data_cum[:, i] - widths
        
        ax.barh(labels, widths, left=starts, height=0.8,
                label=colname, color=color)
        xcenters = starts + widths / 2
 
        text_color = 'black'
        for y, (x, c) in enumerate(zip(xcenters, widths)):
            ax.text(x, y, str(int(c))+ "%", ha='center', va='center',
                    color=text_color, fontsize='large')
    
    ax.set_xlabel("Percentage", fontsize="large")
    ax.set_yticklabels(labels, fontsize="large")
     
    ax.legend(ncol=len(category_names), bbox_to_anchor=(0, 1),
              loc='lower left', fontsize='large')
    
    return
    


# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np

def to_dataframe(selection):
    """convert selection tuples to dataframe
    """
    result = {}

    if len(selection["rerouted"]) == 0:
        result['rerouted'] = pd.DataFrame(columns=['conversation_id','log_id'], dtype='object')
    else:
        result['rerouted'] = pd.DataFrame({'conversation_id': np.array(list(zip(*selection["rerouted"]))[0]), 'log_id': np.array(list(zip(*selection["rerouted"]))[1])})

    if len(selection["dropped_off"]) == 0:
        result['dropped_off'] = pd.DataFrame(columns=['conversation_id','log_id'], dtype='object')
    else:
        result['dropped_off'] = pd.DataFrame({'conversation_id': np.array(list(zip(*selection["dropped_off"]))[0]), 'log_id': np.array(list(zip(*selection["dropped_off"]))[1])})
    #return pd.DataFrame({'conversation_id': np.array(list(zip(*selection))[0]), 'log_id': np.array(list(zip(*selection))[1])})
    return result

def extract_conversation_transcript(df_logs, selection_df, index):
    """extract conversation transcript from logs, based on selection df, and index of interest
    """
    selected_by_index = selection_df.iloc[index]
    transcript = df_logs[df_logs['conversation_id'] == selected_by_index['conversation_id']]
    transcript = transcript.sort_values(by='response_timestamp')
    transcript['you_are_here'] = ""
    transcript.loc[transcript['log_id'] == selected_by_index['log_id'], 'you_are_here'] = '--->'
    return transcript[['you_are_here','response_timestamp','request_text','response_text','log_id','conversation_id','nodes_visited', 'node_visited',
       'branch_exited', 'branch_exited_reason']]


def fetch_logs(logs_df, selection_df):
    """fetch all logs from conversations in selection_df
    """
    print("Deprecated.  Use fetch_logs_by_selection or fetch_logs_by_id")
    return logs_df[logs_df['conversation_id'].isin(list(selection_df['conversation_id']))]

def fetch_logs_by_selection(logs_df, selection_df):
    """fetch all logs from conversations in selection_df
    """
    return logs_df[logs_df['conversation_id'].isin(list(selection_df['conversation_id']))]
    
def fetch_logs_by_id(logs_df, conversation_id):
    """fetch logs by conversation id
    """
    return logs_df[logs_df['conversation_id'] == conversation_id] 

##############################################################################
# fetching utterances from pathflow or milestone chart selection
def get_all_utterances_from_selection(selection, df_logs):
    
    selection_df = to_dataframe(selection)
 
    dropped_off_df = selection_df["dropped_off"]

    dropped_off_conversations = fetch_logs_by_selection(df_logs, dropped_off_df)
    
    input_list=dropped_off_conversations[dropped_off_conversations.request_text!=""].request_text.tolist()
    
    return input_list

def get_last_utterances_from_selection(selection, df_logs):
    
    selection_df = to_dataframe(selection)
 
    dropped_off_df = selection_df["dropped_off"]

    dropped_off_conversations = fetch_logs_by_selection(df_logs, dropped_off_df).sort_values(by=["conversation_id","response_timestamp"])  
    
    input_list=[]
    
    previous_conversation_id=""
    
    for i in range(0,len(dropped_off_conversations)):         
         
        current_conversation_id = dropped_off_conversations.iloc[i,dropped_off_conversations.columns.get_loc("conversation_id")]
        
        if current_conversation_id != previous_conversation_id and i>0:
            input_list.append(dropped_off_conversations.iloc[i-1,dropped_off_conversations.columns.get_loc("request_text")])
                 
        previous_conversation_id = current_conversation_id 
     
    input_list.append(dropped_off_conversations.iloc[len(dropped_off_conversations)-1,dropped_off_conversations.columns.get_loc("request_text")])
           
    return input_list


# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nltk.corpus import stopwords  
#####################################################################

##############################################################
import re 
import pandas as pd

#######################################################################
from sklearn.feature_extraction.text import CountVectorizer 
 

def order_ngram(ngram): 
    words = ngram.split()
    # sort the list.
    words.sort()
    return " ".join(words)

##############################################################
# returns clean string
def clean_text(text, custom_stop_words = []):
    
    REPLACE_BY_SPACE_RE = re.compile('[/(){}\[\]\|@,;]')
    BAD_SYMBOLS_RE = re.compile('[^0-9a-z #+_:]') 
    text = text.lower() # lowercase text
    text = REPLACE_BY_SPACE_RE.sub(' ', text) # replace REPLACE_BY_SPACE_RE symbols by space in text
    text = BAD_SYMBOLS_RE.sub('', text) # delete symbols which are in BAD_SYMBOLS_RE from text
    
    STOPWORDS = set(stopwords.words('english'))
    STOPWORDS = STOPWORDS | set(custom_stop_words)
    
    text = ' '.join(word for word in text.split() if word not in STOPWORDS) 
    
    return text
 
def get_frequent_words_bigrams(utterances, num_unigrams,num_bigrams,custom_stop_words):
    
    utterances = [clean_text(x,custom_stop_words) for x in utterances]
    
    if len(utterances)==0:
        print("Warning! List of utterances is empty")
        return {"name": "","children": []}
    
    unigrams_found = False
    bigrams_found = False
    
    for i in range(0, len(utterances)):
        current_length=len(utterances[i].split())
        if current_length>1:
            unigrams_found = True
            bigrams_found = True
            break
        elif current_length>0:
            unigrams_found = True
    
    if not unigrams_found:
        print("Warning! List of utterances is empty. Perhaps the documents only contain stop words")
        return {"name": "","children": []}
    
    cv_words = CountVectorizer(ngram_range=(1,1))
    
    cv_fit_words=cv_words.fit_transform(utterances)     
    word_list = cv_words.get_feature_names();    
    word_count_list  = cv_fit_words.toarray().sum(axis=0)    
    words_zip = sorted(zip(word_count_list, word_list),reverse=True)
    words_zip=words_zip[0:min(len(words_zip), num_unigrams)]
    
    if bigrams_found:
        cv_bigrams = CountVectorizer(ngram_range=(2,2))
        
        cv_fit_bigrams=cv_bigrams.fit_transform(utterances)     
        bigram_list = cv_bigrams.get_feature_names();    
        bigram_count_list  = cv_fit_bigrams.toarray().sum(axis=0)    
        bigram_zip = sorted(zip(bigram_count_list, bigram_list),reverse=True)
        bigram_zip=bigram_zip[0:num_bigrams]
        
    else:
        print("Warning! List of bigrams is empty.")

    children = []
    for i in range(0, min(len(words_zip), num_unigrams)):
        children.append({"name": words_zip[i][1], "value": int(words_zip[i][0])})
    if bigrams_found:
        for i in range(0, min(len(bigram_zip), num_bigrams)):
            children.append({"name": bigram_zip[i][1], "value": int(bigram_zip[i][0])})
    data = {"name": "","children": children}

    return data
 
##########################################################################################################
def get_df_frequencies(neg_dict, pos_dict, neg_list, pos_list, table_title1, table_title2):
    neg_list_neg_freq=[]
    neg_list_pos_freq=[]
    pos_list_neg_freq=[]
    pos_list_pos_freq=[]
     
    num_neg_features = len(neg_list)
    if num_neg_features==0:
        print("Warning! No significant keywords were found.")
        
    num_pos_features = len(pos_list)
    
    for i in range(0, num_neg_features):
        neg_list_neg_freq.append(neg_dict[neg_list[i]][0])              
        neg_list_pos_freq.append(neg_dict[neg_list[i]][2]) 
        
    for i in range(0, num_pos_features):        
        pos_list_pos_freq.append(pos_dict[pos_list[i]][0])              
        pos_list_neg_freq.append(pos_dict[pos_list[i]][2])
                
    if num_neg_features>num_pos_features:
        D = [None]*(num_neg_features-num_pos_features)
        pos_list.extend(D)
        pos_list_pos_freq.extend(D)
        pos_list_neg_freq.extend(D)
    
    if num_pos_features>num_neg_features:
        D = [None]*(num_pos_features-num_neg_features)
        neg_list.extend(D)
        neg_list_neg_freq.extend(D)
        neg_list_pos_freq.extend(D)
    
    df_frequencies = pd.DataFrame() 
    df_frequencies[table_title1] = neg_list 
    df_frequencies["Failure keywords: negative corpus frequencies, %"] = neg_list_neg_freq
    df_frequencies["Failure keywords: positive corpus frequencies, %"] = neg_list_pos_freq
    df_frequencies[table_title2] = pos_list
    df_frequencies["Success keywords: positive corpus frequencies, %"] = pos_list_pos_freq
    df_frequencies["Success keywords: negative corpus frequencies, %"] = pos_list_neg_freq

    return df_frequencies    


########################################################################
def get_features(text_negative, text_positive, num_features, n_range=(1,2)):    
    
    PRINT_MODE = False
    
    #no_shows=["account","number"]
    no_shows=[]
    
    ratio_parameter=2.0
    num_negative=len(text_negative)
    num_positive=len(text_positive)
     
    cv_negative = CountVectorizer(ngram_range=n_range)
    
    cv_fit=cv_negative.fit_transform(text_negative)     
    word_list_negative = cv_negative.get_feature_names();    
    count_list_negative = cv_fit.toarray().sum(axis=0)    
    negative_count_unigrams = sorted(zip(count_list_negative, word_list_negative),reverse=True)  
    dict_negative = {word_list_negative[i]: count_list_negative[i] for i in range(len(word_list_negative))} 
 
    cv_positive = CountVectorizer(ngram_range=n_range)
             
    cv_fit=cv_positive.fit_transform(text_positive)     
    word_list_positive = cv_positive.get_feature_names();    
    count_list_positive = cv_fit.toarray().sum(axis=0)    
    positive_count_unigrams = sorted(zip(count_list_positive, word_list_positive),reverse=True)  
    dict_positive = {word_list_positive[i]: count_list_positive[i] for i in range(len(word_list_positive))} 
    
    ##############################
    # perform frequence check, print warnings
    counter_neg_features=0
    counter=0
    neg_list_sorted=[]
    neg_list=[]
    neg_dict={}
    while counter_neg_features<num_features:
        current_neg_feature=negative_count_unigrams[counter][1] 
        current_neg_count=negative_count_unigrams[counter][0]
            
        get_neg_freq=current_neg_count / num_negative 
        
        if current_neg_feature in dict_positive.keys():
            current_pos_count = dict_positive[current_neg_feature]
            get_pos_freq = (current_pos_count + 0.0) / num_positive 
        else:
            current_pos_count = 0
            get_pos_freq = 1.0 / num_positive 
         
        if ratio_parameter*get_pos_freq>=get_neg_freq:
            if PRINT_MODE:
                print("Frequencies for negative feature "+current_neg_feature+" are not appropriate. Feature is omitted")
                print("Negative frequency: "+str(get_neg_freq))
                print("Positive frequency: "+str(get_pos_freq))
                
        elif current_neg_feature not in no_shows:
            S = order_ngram(current_neg_feature)
            if S in neg_list_sorted:
                ind = neg_list_sorted.index(S)
                prev = neg_list[ind] 
                neg_dict[prev][0] += current_neg_count
                neg_dict[prev][1] += current_pos_count
                neg_dict[prev][2] += ((current_neg_count + neg_dict[prev][0]) / max(current_pos_count + neg_dict[prev][1],1)) *\
                    (num_positive/num_negative)
            else:
                neg_list.append(current_neg_feature)
                neg_list_sorted.append(S)

                neg_dict[current_neg_feature] = [current_neg_count, current_pos_count, get_neg_freq / get_pos_freq]
                counter_neg_features+=1
        
        counter+=1 
        
        if counter==len(negative_count_unigrams):
            break
        
    counter_pos_features=0
    counter=0
    pos_list_sorted=[]
    pos_list=[]
    pos_dict={}
    while counter_pos_features<num_features:
        current_pos_feature = positive_count_unigrams[counter][1] 
        current_pos_count = positive_count_unigrams[counter][0]
        
        get_pos_freq=current_pos_count / num_positive
        
        if current_pos_feature in dict_negative.keys():
            current_neg_count = dict_negative[current_pos_feature]
            get_neg_freq = dict_negative[current_pos_feature]
        else:
            current_neg_count = 0
            get_neg_freq = 1.0 / num_negative         
        
        if get_pos_freq<=ratio_parameter*get_neg_freq:
            if PRINT_MODE:
                print("Frequencies for positive feature '"+current_pos_feature+"' are not appropriate. Feature is omitted.")
                print("Negative frequency: "+str(get_neg_freq))
                print("Positive frequency: "+str(get_pos_freq))
        elif current_pos_feature not in no_shows:
            S = order_ngram(current_pos_feature)
            if S in pos_list_sorted:
                ind = pos_list_sorted.index(S)
                prev = pos_list[ind] 
                pos_dict[prev][0] += current_pos_count
                pos_dict[prev][1] += current_neg_count
                pos_dict[prev][2] += ((current_pos_count + pos_dict[prev][0]) / max(current_neg_count + pos_dict[prev][1],1)) *\
                    (num_negative/num_positive )
            else:
                pos_list.append(current_pos_feature)
                pos_list_sorted.append(S)
                
                pos_dict[current_pos_feature] = [current_pos_count, current_neg_count, get_pos_freq / get_neg_freq]
                counter_pos_features+=1
        
        counter+=1    
        
        if counter==len(positive_count_unigrams):
            break  
    
    return neg_dict, pos_dict, neg_list, pos_list 



########################################################################
def keyword_table_detailed(text_negative, text_positive, num_keywords, custom_stop_words, title1="Title 1", title2="Title 2"):
    
    COEFF = [0.4, 0.4, 0.2]
    
    text_positive = [clean_text(x,custom_stop_words) for x in text_positive]
    text_negative = [clean_text(x,custom_stop_words) for x in text_negative]
    
    if len(text_positive)==0 or len(text_negative)==0:
        print("Warning! At least one of List of utterances is empty")
        neg_dict = {}
        pos_dict = {}
        neg_list = [] 
        pos_list = []
        
    else:        
        
        unigrams_found_negative = False
        bigrams_found_negative = False
        trigrams_found_negative = False
        
        for i in range(0, len(text_negative)):
            current_length=len(text_negative[i].split())
            if current_length>2:
                unigrams_found_negative = True
                bigrams_found_negative = True
                trigrams_found_negative = True
                break
            elif current_length>1:
                unigrams_found_negative = True
                bigrams_found_negative = True
            elif current_length>0:
                unigrams_found_negative = True
        
        ################################################################        
        unigrams_found_positive = False
        bigrams_found_positive = False
        trigrams_found_positive = False
        
        for i in range(0, len(text_positive)):
            current_length=len(text_positive[i].split())
            if current_length>2:
                unigrams_found_positive = True
                bigrams_found_positive = True
                trigrams_found_positive = True
                break
            elif current_length>1:
                unigrams_found_positive = True
                bigrams_found_positive = True
            elif current_length>0:
                unigrams_found_positive = True        
        
        ####################################################################
        unigrams_found = unigrams_found_negative and unigrams_found_positive
        bigrams_found = bigrams_found_negative and bigrams_found_positive
        trigrams_found = trigrams_found_negative and trigrams_found_positive
        
        if not unigrams_found:
            print("Warning! At least one of utterance lists is empty. Perhaps the documents only contain stop words")
            neg_dict = {}
            pos_dict = {}
            neg_list = [] 
            pos_list = []
        else:
            num_keywords1 = round(num_keywords*COEFF[0])
        
            neg_dict, pos_dict, neg_list, pos_list =  get_features(text_negative,text_positive,num_keywords1,n_range=(1,1))
            
            if not bigrams_found:
                print("Warning! No bigrams for at least one of utterance lists.")  
            else:
     
                num_keywords2 = round(num_keywords*COEFF[1])
            
                neg_dict2, pos_dict2, neg_list2, pos_list2 =  get_features(text_negative,text_positive,num_keywords2,n_range=(2,2))
                
                neg_list.extend(neg_list2)  
                pos_list.extend(pos_list2) 
                
                neg_dict.update(neg_dict2) 
                pos_dict.update(pos_dict2) 
                
                if not trigrams_found:
                    print("Warning! No trigrams for at least one of utterance lists.")    
                else:  
                    num_keywords3 = num_keywords - num_keywords1 - num_keywords2 
    
                    neg_dict3, pos_dict3, neg_list3, pos_list3 =  get_features(text_negative,text_positive,num_keywords3,n_range=(3,3))
     
                    neg_list.extend(neg_list3)                      
                    pos_list.extend(pos_list3)                     
                   
                    neg_dict.update(neg_dict3)                   
                    pos_dict.update(pos_dict3)
  
    ####################################################################################################   
    df_frequencies = get_df_frequencies(neg_dict, pos_dict, neg_list, pos_list, title1, title2) 
    
    df_frequencies.rename(columns={"Failure keywords: negative corpus frequencies, %": title1+": Frequency",\
                                   "Failure keywords: positive corpus frequencies, %": title1+": Power",\
                                   "Success keywords: positive corpus frequencies, %": title2+": Frequency",\
                                   "Success keywords: negative corpus frequencies, %": title2+": Power"}, inplace=True)
     
    return df_frequencies

################################################################################################
def get_data_for_comparison_visual(user_input_abandoned, user_input_completed, num_keywords, custom_stop_words=[]):
        
    df_keywords_detailed = keyword_table_detailed(user_input_abandoned, user_input_completed, num_keywords, custom_stop_words)
    
    #print(df_keywords_detailed.head(n=25))
    
    data={}
    data["name"]=""
    data["children"]=[]
    for i in range(0,len(df_keywords_detailed)):
        if  df_keywords_detailed.iloc[i,1] is None:
            data["children"].append({"name": df_keywords_detailed.iloc[i,0], "value": df_keywords_detailed.iloc[i,1]})
        else:
            data["children"].append({"name": df_keywords_detailed.iloc[i,0], "value": int(df_keywords_detailed.iloc[i,1])})
        
    return data  


# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from string import punctuation
EN_PUNCTUATION = punctuation + '’'

def sanitize_text(text, remove_punctuation=True, lower=True, tokenize=True):
    text = text.strip()
    if lower:
        text = text.lower()
    # if tokenize:
    #     words = word_tokenize(text)
    # else:
    #     words = text.split()
    # if remove_punctuation:
    #     words = [word for word in words if word not in EN_PUNCTUATION]
    # return ' '.join(words)
    if remove_punctuation:
        text = text.translate(str.maketrans('', '', EN_PUNCTUATION))
    return text

def filter_sentences(sentences, sentence_with_count=None, min_complexity=-1, max_complexity=-1, min_confidence=-1,
                     max_confidence=-1, max_sentences_limit = None):
    result = []
    # apply filter based on complexity and confidence
    discarded = 0
    for index, sentence in sentences.iterrows():
        if min_complexity > 0 or max_complexity > 0:
            clean_sentence = sanitize_text(sentence['example'])
            complexity = len(clean_sentence.split())
            if min_complexity > 0 and complexity < min_complexity:
                discarded += 1
                continue
            if 0 < max_complexity < complexity:
                discarded += 1
                continue
        if min_confidence > 0:
            if sentence['confidence'] < min_confidence:
                discarded += 1
                continue
        if max_confidence > 0:
            if sentence['confidence'] > max_confidence:
                discarded += 1
                continue
        if sentence_with_count is not None:
            if sentence['example'] in sentence_with_count:
                sentence_with_count[sentence['example']] += 1
                discarded += 1
                continue

        result.append(sentence['example'])

        if max_sentences_limit:
            if len(result) == max_sentences_limit:
                logger.info("\tmax sentences limit reached. We will not consider any remaining examples/sentences")
                break

    return result



# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

###################################################################
from textblob import TextBlob
import re

########################################################################
def add_sentiment_columns(df_logs_canonical):
    df_logs_canonical['sentiment_result']=\
        [utterance_sentiment_detailed(user_input) for user_input in df_logs_canonical['request_text']]
    
    df_logs_canonical['sentiment']=df_logs_canonical['sentiment_result'].str[0]        
    df_logs_canonical['sentiment_positive_part']=df_logs_canonical['sentiment_result'].str[1]  
    df_logs_canonical['sentiment_negative_part']=df_logs_canonical['sentiment_result'].str[2]  
    
    del df_logs_canonical['sentiment_result']
    
    return df_logs_canonical

#################################################
def is_thanks(str):

    list_thanks =['thank']

    list_words=str.split()
    num_words=len(list_words)

    if num_words<5:

        for i in range(0,num_words):
            current_word=list_words[i].lower()
            current_word=re.sub("[^a-zA-Z]+", "",current_word)
            if current_word in list_thanks:
                return True, current_word

    return False, ""

########################################################

def utterance_sentiment_detailed(user_input):
    
    sentiment = TextBlob(user_input)
    
    current_blob_polarity=sentiment.sentiment.polarity
    current_blob_assessment = sentiment.sentiment_assessments.assessments
    
    num_assessments =len(current_blob_assessment)
    
    positive_list=""
    negative_list=""
    
    for i in range(0,num_assessments):
        current_sent = current_blob_assessment[i][1]
        current_words = current_blob_assessment[i][0]
        num_words=len(current_words)
        
        if current_sent>0:
            for j in range(0, num_words):
                positive_list += (current_words[j]+" ")
        elif current_sent<0:
            for j in range(0, num_words):
                negative_list += (current_words[j]+" ")           
    
    if current_blob_polarity==0:
        is_th = is_thanks(user_input)
        if is_th[0]: 
            current_blob_polarity=0.2
            positive_list += (is_th[1]+" ")
    
    return current_blob_polarity, positive_list, negative_list

###########################################################

def utterance_sentiment(user_input):
    
    sentiment = TextBlob(user_input)
    
    current_blob_estimate=sentiment.sentiment.polarity
    
    if current_blob_estimate==0:
        if is_thanks(user_input):
            current_blob_estimate=0.5
    
    return current_blob_estimate

#############################################

def normalize_sentiment(sentiment):
    
    if sentiment<-1:
        return -1
    elif sentiment>1:
        return 1
    else: 
        return sentiment     



# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
#import requests
from IPython.display import display, Javascript, HTML
import json
import datetime
import os
import os.path
import pkgutil
import inspect
import conversation_analytics_toolkit as cat

def read_file(filename):    
    with open(filename, 'r') as file:
        data = file.read().replace('<style>', '').replace('</style>', '')
        #print(data)
    return data

def add_style_to_header(css_string):
    display(Javascript("""
        var newStyle = document.createElement("style");
        newStyle.innerHTML = `%s`;
        document.getElementsByTagName("head")[0].appendChild(newStyle);
    """ % (css_string)))

if not cat.is_production() and not cat.is_forced_production_mode():
    project_folder = cat._get_src_project_folder()
    
    #add CSS to header instead of in body cell (prevent styling from disappearing even if output cell is deleted)
    add_style_to_header(read_file(os.path.join(project_folder, "src", "conversation_analytics_toolkit", "flowchart.css")))
    add_style_to_header(read_file(os.path.join(project_folder, "src", "conversation_analytics_toolkit", "transcript.css")))
    add_style_to_header(read_file(os.path.join(project_folder, "src", "conversation_analytics_toolkit", "wordpackchart.css")))
    add_style_to_header(read_file(os.path.join(project_folder, "src", "conversation_analytics_toolkit", "wa_dialog_chart.css")))

    display(Javascript("require.config({paths: {d3: 'https://cdnjs.cloudflare.com/ajax/libs/d3/4.13.0/d3.min'}});"))
    display(Javascript("require.config({paths: {lodash: 'https://cdn.jsdelivr.net/npm/lodash@4.17.15/lodash.min'}});"))
    display(Javascript(filename=os.path.join(project_folder, "src","conversation_analytics_toolkit", "flowchart2.js")))
    display(Javascript(filename=os.path.join(project_folder, "src","conversation_analytics_toolkit", "transcript.js")))
    display(Javascript(filename=os.path.join(project_folder, "src","conversation_analytics_toolkit", "wordpackchart.js")))   
   
    display(Javascript(filename=os.path.join(project_folder, "src", "conversation_analytics_toolkit", "wa_model.js")))
    display(Javascript(filename=os.path.join(project_folder, "src", "conversation_analytics_toolkit", "wa_tree.js")))
    display(Javascript(filename=os.path.join(project_folder, "src", "conversation_analytics_toolkit", "wa_node_details.js")))
    display(Javascript(filename=os.path.join(project_folder, "src", "conversation_analytics_toolkit", "wa_dialog_chart.js")))
else:
    flowchart_css = pkgutil.get_data(__package__, 'flowchart.min.css').decode('utf-8')
    flowchart_js = pkgutil.get_data(__package__, 'flowchart2.min.js').decode('utf-8')
    transcript_css = pkgutil.get_data(__package__, 'transcript.min.css').decode('utf-8')
    transcript_js = pkgutil.get_data(__package__, 'transcript.min.js').decode('utf-8')
    wordpackchart_css = pkgutil.get_data(__package__, 'wordpackchart.min.css').decode('utf-8')
    wordpackchart_js = pkgutil.get_data(__package__, 'wordpackchart.min.js').decode('utf-8')

    wa_model_js = pkgutil.get_data(__package__, 'wa_model.min.js').decode('utf-8')
    wa_tree_js = pkgutil.get_data(__package__, 'wa_tree.min.js').decode('utf-8')
    wa_node_details_js =  pkgutil.get_data(__package__, "wa_node_details.min.js").decode('utf-8')
    wa_dialog_chart_js = pkgutil.get_data(__package__, 'wa_dialog_chart.min.js').decode('utf-8')
    wa_dialog_chart_css = pkgutil.get_data(__package__, 'wa_dialog_chart.min.css').decode('utf-8')

    #add CSS to header instead of in body cell (prevent styling from disappearing even if output cell is deleted)
    add_style_to_header(flowchart_css)
    add_style_to_header(transcript_css)
    add_style_to_header(wordpackchart_css)
    add_style_to_header(wa_dialog_chart_css)
    
    display(Javascript("require.config({paths: {d3: 'https://cdnjs.cloudflare.com/ajax/libs/d3/4.13.0/d3.min'}});"))
    display(Javascript("require.config({paths: {lodash: 'https://cdn.jsdelivr.net/npm/lodash@4.17.15/lodash.min'}});"))
    display(Javascript(data=flowchart_js))
    display(Javascript(data=transcript_js))
    display(Javascript(data=wordpackchart_js))
    display(Javascript(data=wa_model_js))
    display(Javascript(data=wa_tree_js))
    display(Javascript(data=wa_node_details_js))
    display(Javascript(data=wa_dialog_chart_js))

def draw_flowchart(config, json_data, python_selection_var="selection", width=600, height=400):
    display(Javascript("""
        (function(element){
            require(['flowchart2'], function(flowchart2) {
                var config = %s;
                if (config["debugger"]===true){
                   debugger;
                };
                var chart = flowchart2(element.get(0), config, %s);

                chart.on("export",function(e){
                    var selection = JSON.stringify(e.selection).replace(/"/g , "'");
                    //var selection = JSON.stringify(e).replace(/"/g , "'");
                    IPython.notebook.kernel.execute("%s = " + selection);
                });
            });
        })(element);
    """ % (json.dumps(config), json.dumps(json_data), python_selection_var)))


def convert_timestamp(item_date_object):
    if isinstance(item_date_object, (datetime.date, datetime.datetime)):
        return item_date_object.isoformat()

def draw_transcript(config, json_data, width=600, height=400):
    display(Javascript("""
        (function(element){
            require(['transcript'], function(transcript) {
                var config = %s;
                if (config["debugger"]===true){
                   debugger;
                };
                var chart = transcript(element.get(0), config, %s);
            });
        })(element);
    """ % (json.dumps(config), json.dumps(json_data, default=convert_timestamp))))
    
def draw_wordpackchart(config, json_data, width=600, height=400):
    display(Javascript("""
        (function(element){
            require(['wordpackchart'], function(wordpackchart) {
                var config = %s;
                var chart = wordpackchart(element.get(0), config, %s);
            });
        })(element);
    """ % (json.dumps(config), json.dumps(json_data, default=convert_timestamp))))

def draw_wa_dialog_chart(config, json_data, width=600, height=400):
    display(Javascript("""
        (function(element){
            require(['wa_dialog_chart'], function(wa_dialog_chart) {
                var config = %s;
                var chart = wa_dialog_chart(element.get(0), config, %s);
            });
        })(element);
    """ % (json.dumps(config), json.dumps(json_data))))


# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json

def _wa_to_ba_json(row, workspace_name, conversation_id_field):
    row_request = row["request"]
    row_response = row["response"]

    ba_json = {}
    ba_json["event"] = "REQUEST_RESPONSE" 
     
    if len(row.request_timestamp)<23:
        timestring = row.request_timestamp[:19]
        ba_json["timestamp"] = int(datetime.datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S").timestamp()*1000)
    else:
        timestring = row.request_timestamp[:23]
        ba_json["timestamp"] = int(datetime.datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S.%f").timestamp()*1000)

    ba_json["log_id"]=row["log_id"]

    ba_json["data"] = {}
    ba_json["data"]["input"] = { "text": row_request["input"]["text"] }
    ba_json["data"]["intents"] = row_response["intents"]
    ba_json["data"]["entities"] = row_response["entities"]
    ba_json["data"]["output"] = {
        "text" : row_response["output"]["text"],
        "nodes_visited": row_response["output"]["nodes_visited"]
    }

    if 'branch_exited' in row_response["context"]["system"]:
        ba_json["data"]["output"]["branch_exited"] = row_response["context"]["system"]["branch_exited"]
        ba_json["data"]["output"]["branch_exited_reason"] = row_response["context"]["system"]["branch_exited_reason"]

    ba_json["data"]["context"] = {}
    ba_json["data"]["context"]["conversation_id"] = row_response["context"][conversation_id_field]
    if "user_id" in row_response["context"]:
        ba_json["data"]["context"]["user_id"] = row_response["context"]["user_id"]
    if "workspace_name" in row_response["context"]:
        ba_json["data"]["context"]["workspace_name"] = row_response["context"]["workspace_name"]
    elif workspace_name != None:
        ba_json["data"]["context"]["workspace_name"] = workspace_name
    if "not_handled" in row_response["context"]:
        # optional. Boolean. Default value is false.
        ba_json["data"]["context"]["not_handled"] = row_response["context"]["not_handled"]

    ba_json["data"]["context"]["input_type"] = "text"
    return ba_json

# create row for sql table
def _wa_to_ba_transformer(row, tenant_id, workspace_name, conversation_id_field):
    ba_row=[]

    ba_row.append(row["log_id"])
    event_json = _wa_to_ba_json(row, workspace_name, conversation_id_field)
    ba_row.append(json.dumps(event_json))
    ba_row.append(int(event_json["timestamp"]))
    ba_row.append(tenant_id)

    return ba_row

def transform_wa_to_ba(df_logs, tenant_id, workspace_name=None, conversation_id_field="conversation_id"):
    """
    Transforms WA log dataframe format to BA array format
    """
    # iterate to get list of rows for sql table
    ba_array = []
    for index, row in df_logs.iterrows():
        ba_array.append(_wa_to_ba_transformer(row, tenant_id, workspace_name, conversation_id_field))
    return ba_array

# store ba_array to the database
def store_ba_to_mysql(ba_array, db_credentials, tenant_id, raw_data_table='analyticsRecords'):
    """
    Persist ba array to MySQL
    """
    import pymysql
    # establish a connection
    db_connection = pymysql.connect(
        host=db_credentials["host"],
        port=db_credentials["port"],
        user=db_credentials["user"],
        password=db_credentials["password"],
        db=db_credentials["database"],
        charset="utf8")

    # insert data to sql table
    try:
        with db_connection.cursor() as db_cursor:
            field_list="(id, event_json, epoch, tenant_id)"
            insert_stmt = "INSERT INTO "+raw_data_table+" "+field_list+" VALUES (%s,%s,%s,%s)"
            db_cursor.executemany(insert_stmt,ba_array)
    except Exception as e:
            print(e)
            print("Failure in database connection")

    db_connection.commit()
    db_connection.close()


def store_ba_to_postgres(ba_array, db_credentials, tenant_id, raw_data_table='analyticsRecords'):
    """
    Persist ba array to MySQL
    """

    import psycopg2
    # establish a connection
    #db_credentials['dbname'] = db_credentials["database"]
    #del db_credentials["database"]
    if db_credentials.setdefault('ssl', False): # setdefault() - Returns the value of the specified key. If the key does not exist: insert the key, with the specified value
        db_credentials['sslmode']='require'
    del db_credentials['ssl']

    db_connection = psycopg2.connect(**db_credentials)

    # insert data to sql table
    try:
        with db_connection.cursor() as db_cursor:
            field_list="(id, event_json, epoch, tenant_id)"
            insert_stmt = "INSERT INTO "+raw_data_table+" "+field_list+" VALUES (%s,%s,%s,%s)"
            #db_cursor.executemany(insert_stmt,ba_array)
            psycopg2.extras.execute_batch(db_cursor, insert_stmt, ba_array, page_size=1000)

    except Exception as e:
            print(e)
            print("Failure in database connection")

    db_connection.commit()
    db_connection.close()



# -*- coding: utf8 -*-
#
# This file were created by Python Boilerplate. Use Python Boilerplate to start
# simple, usable and best-practices compliant Python projects.
#
# Learn more about it at: http://github.com/fabiommendes/python-boilerplate/
#

import setuptools
import re

__version__ = '1.0.5'

# Convert README.md to README.rst for pypi
try:
    from pypandoc import convert_file

    def read_md(f):
        return re.sub(r'\nContributor List(.*)', '', convert_file(f, 'rst'), flags=re.S)

except:
    print('warning: pypandoc module not found, '
          'could not convert Markdown to RST')

    def read_md(f):
        return open(f, 'rb').read().decode(encoding='utf-8')

setuptools.setup(
    # Basic info
    name='conversation_analytics_toolkit',
    version=__version__,
    author='IBM Watson',
    author_email='watdevex@us.ibm.com',
    maintainer='Avi Yaeli',
    maintainer_email='aviy@il.ibm.com',
    url='https://github.com/watson-developer-cloud/assistant-dialog-flow-analysis',
    description='Dialog Flow Analysis Tool for Watson Assistant',
    license='Apache 2.0',
    long_description=read_md('README.md'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    # Packages and depencies
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    install_requires=[
        'pandas~=0.24.0',
        'nltk>=3.4.0',
        'textblob>=0.15.3',
        'scikit-learn>=0.21.3',
        'tqdm>=4.31.1',
        'ipywidgets>=7.4.2',
        'ibm-watson>=4.3.0',
        'plotly>=4.5.0',
        'requests>=2.18.4'
    ],

    # extras_require={
    #     'dev': [
    #     ]
    # },

    # Data files
    package_data={
        'conversation_analytics_toolkit': [
            #'images/*.*',
            '*.min.css',
            '*.min.js'
        ],
    },

    zip_safe=False,
    platforms='any',
)



# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import pandas as pd
import numpy as np
import json
import conversation_analytics_toolkit as cat
from pandas.io.json import json_normalize

class _conversationPathAnalysis:
    def __init__(self, mode, on_column, max_depth, trim_reroutes):
        self.mode = mode
        self.on_column = on_column
        self.max_depth = max_depth
        self.trim_reroutes = trim_reroutes
        self.nodes_paths = defaultdict(list)

    def handle_node_row(self, path, row, step, is_conversation_start, path_length):
        if not path in self.nodes_paths:
            self.nodes_paths[path] = {"flows":0, "rerouted":0, "dropped_off":0, "type": "NODE", "name": row[self.on_column],"is_conversation_start": is_conversation_start, "conversation_log_ids_dropped_off":[], "conversation_log_ids_rerouted":[],"path_length": path_length}
        self.nodes_paths[path]["flows"] += 1
        #if (row.response_branch_exited==True and row.response_branch_exited_reason=='fallback' and row.digression is not True):
        if self.trim_reroutes == True:
            if (row.branch_exited==True and row.branch_exited_reason=='fallback' and row.digression is False):
                self.nodes_paths[path]["rerouted"] += 1
                self.nodes_paths[path]["conversation_log_ids_rerouted"].append((row.conversation_id, row.log_id))
                return 'rerouted'

    def handle_conversation_node_exit(self, path, step, row):
        self.nodes_paths[path]["dropped_off"] += 1
        self.nodes_paths[path]["conversation_log_ids_dropped_off"].append((row.conversation_id, row.log_id))

    def handle_conversation_path(self, df):
        # if data doesn't include digression, add it for downstream dependency.
        if 'digression' not in df.columns:
            df["digression"] = False

        prev_node_path = ""
        prev_row = ""
        delimiter = "\\"
        for i, (index, row) in enumerate(df.iterrows()):
            node_code = row[self.on_column]
            node_id = str(node_code)
            if i==0:
                curr_node_path = node_id
                path_length = 0
                if self.handle_node_row(curr_node_path, row, i, True, path_length) == 'rerouted':
                    # note this will never happen if trim_reroutes=False
                    break
            else:
                curr_node_path = prev_node_path + delimiter + node_id
                path_length += 1
                if (i <= self.max_depth):
                    if self.handle_node_row(curr_node_path, row, i, True, path_length) == 'rerouted':
                        # note this will never happen if trim_reroutes=False
                        break
                else:
                    print("reached depth limit of " + str(i-1) + ", ignoring steps beyond.  You can set max_depth to a larger number")
                    self.handle_conversation_node_exit(prev_node_path, i-1, row)
                    break
            prev_node_path = curr_node_path
            prev_row = row
            if i == len(df) - 1:
                self.handle_conversation_node_exit(curr_node_path, i, row)

class DialogNodesGraph(): 
    # compute visits in dialog nodes
    def __init__(self, workspace, description=""):
        self.steps = None
        df_workspace = json_normalize(workspace)
        # dataframe representation of the workspace
        self.dialog_nodes = df_workspace['dialog_nodes'].values[0]
        # compute map from node to its index in the dialog_nodes_array
        self.dialog_node_index = {}
        for i, element in enumerate(self.dialog_nodes):
            node_id = element["dialog_node"]
            self.dialog_node_index[node_id] = i
        self.description = ""
        # count visits of nodes_visited in each node, prev, and next
        self.nodes_visited_graph = {}
        # map of conversations and their steps
        self.conversation_steps = {}

    def _get_node(self, node_id):
        return self.dialog_nodes[self.dialog_node_index[node_id]]

    def _add_conversation_visits_to_graph(self, nodes_visited):
        for i, node in enumerate(nodes_visited):
            graph = self.nodes_visited_graph
            # add node if not exist
            if node not in graph:
                graph[node] = {'visits': 0, 'prev': {}, 'next': {}}
            # add count to current node
            graph[node]['visits'] += 1
            # add next to every node except last
            if i < len(nodes_visited) - 1:
                next_node = nodes_visited[i+1]
                if next_node not in graph[node]['next']:
                    graph[node]['next'][next_node] = {'visits': 0}
                graph[node]['next'][next_node]['visits']+=1
            # add prev to every node except first
            if i > 0:
                prev_node = nodes_visited[i-1]
                if prev_node not in graph[node]['prev']:
                    graph[node]['prev'][prev_node] = {'visits': 0}
                graph[node]['prev'][prev_node]['visits']+=1
    
    def _slider_update(self, slider, step, step_desc, step_message = None):
        width = 90
        prefix = ""
        long_message = prefix + step_desc + '.' * (width - len(step_desc) - len(prefix))
        slider.update(step)
        if step_message is not None:
            slider.write(step_message)
        slider.set_description(long_message)

    def compute_visits(self, logs):
        if not type(self.steps) == pd.core.frame.DataFrame:
            self.steps = logs
        else:
            self.steps = self.steps.append(logs)
        conversations_grouped =  logs.groupby("conversation_id")
        from tqdm import tqdm, tqdm_notebook
        if cat.use_widgets():
            slider = tqdm_notebook(range(100), desc="processing {} conversations".format(len(conversations_grouped)), ncols=900)
        else:
            slider = tqdm(range(100), desc="processing {} conversations".format(len(conversations_grouped)), ncols=90)
        i=0
        for conversation, steps_df in conversations_grouped:
            i +=1
            if i/100 == int(i/100): # every 100 conversations update the slider
                step = int(100/len(conversations_grouped)*100)
                self._slider_update(slider, step, "Processed {} of {} conversations".format(i, len(conversations_grouped)))
            nodes_visited_in_conversation = []
            steps_df = steps_df.sort_values(["response_timestamp"])
            for index, row in steps_df.iterrows():
                nodes_visited_in_conversation.extend(row["nodes_visited"]) 
            self.conversation_steps[conversation] = nodes_visited_in_conversation
            self._add_conversation_visits_to_graph(nodes_visited_in_conversation)
        self._slider_update(slider, 100, "Processed {} conversations".format(i))

    def get_conversation_steps(self, conversation_id):
        return self.conversation_steps[conversation_id]

    def was_node_visited(self, node_id):
        return True if node_id in self.nodes_visited_graph else False

    def get_nodes_visited(self, node_id):
        return self.nodes_visited_graph[node_id]

    def get_node(self, node_id):
        return self._get_node(node_id)

    def get_node_label(self, node, show_type = True, show_id_if_title_exist = True, show_visits = True):
        node_type = node["type"] if show_type == True else ""
        node_title = node["title"] if "title" in node else ""
        node_id = node["dialog_node"]
        if self.was_node_visited(node_id):
            node_visits = self.get_nodes_visited(node_id)["visits"]
        else: 
            node_visits = 0
        node_label = ""
        if show_type:
            node_label = "[{}]".format(node_type)
        if not show_id_if_title_exist and node_title != "":
            node_label = node_label + " " + node_title
        else: 
            node_label = node_label + " " + node_title + " : " + node_id
        if show_visits == True:
            node_label = node_label + " [ {} ]".format(node_visits)
        return node_label

    def info(self):
        return {
            'converstions': len(self.conversation_steps.keys()),
            'steps': len(self.steps),
            'nodes': len(self.nodes_visited_graph.keys())
        }

    def flows_to_from_node_sankey(self, node_id, arrangement="freeform", orientation='h', pad=10, hovermode='x'):
        import plotly.graph_objects as go 
        current_node_index = self.dialog_node_index[node_id]
        nodes_label = [self.get_node_label(n) for n in self.dialog_nodes]
        sankey_nodes = { "label": nodes_label, "pad": 10}
        #get links for the given node from the graph
        node_graph = self.get_nodes_visited(node_id)
        sankey_links = {"source": [], "target": [], "value": []}
        #iterate on prev
        for prev_node, prev_value in node_graph["prev"].items():
            sankey_links["source"].append(self.dialog_node_index[prev_node])
            sankey_links["target"].append(current_node_index)
            sankey_links["value"].append(prev_value["visits"])
        for next_node, next_value in node_graph["next"].items():
            sankey_links["source"].append(current_node_index)
            sankey_links["target"].append(self.dialog_node_index[next_node])
            sankey_links["value"].append(next_value["visits"])
        
        fig = go.Figure(go.Sankey(
            valuesuffix = " visits",
            orientation = orientation,
            arrangement = arrangement,
            node = {
                "label": sankey_nodes["label"],
                'pad':pad},
            link = {
                "source": sankey_links["source"],

                "target": sankey_links["target"],
                "value": sankey_links["value"]}))
        fig.update_layout(title_text="Nodes visited for node: " + node_id , font_size=10, hovermode = hovermode)
        fig.show()
        return 

    def flows_to_from_node_html(self, node_id):
        node_graph = self.get_nodes_visited(node_id)
        prev_list = []
        for prev_node, prev_value in node_graph["prev"].items():
            prev_list.append({"prev_node": prev_node, "type": self.get_node(node_id)["type"], "visits": prev_value["visits"]})
        prev_df = pd.DataFrame(prev_list)
        next_list = []
        for next_node, next_value in node_graph["next"].items():
            next_list.append({"next_node": next_node, "type": self.get_node(node_id)["type"], "visits": next_value["visits"]})
        next_df = pd.DataFrame(next_list)
        from IPython.display import display, HTML
        display(HTML(prev_df.to_html()))
        display(HTML(next_df.to_html()))

class MilestoneFlowGraph: 
    # data structures
    # { milestones: m1 --> {}}
    # { funnels: funnels --> [m1, m2]} 
    # { mappings: node --> milestone}
    # compute visits in dialog nodes
    def __init__(self, workspace):
        #map of milestones
        self.milestones = {}
        #map of funnels, initialized default
        self.funnels = {
            "default": []
        }
        #nodes mapping to milestones
        self.nodes  = {}
        df_workspace = json_normalize(workspace)
        self.dialog_nodes = df_workspace['dialog_nodes'].values[0]
        # compute map from node to its index in the dialog_nodes_array
        self.dialog_node_index = {}
        for i, element in enumerate(self.dialog_nodes):
            node_id = element["dialog_node"]
            self.dialog_node_index[node_id] = i

    def add_milestones(self, names, funnel="default", atIndex=None):
        # add milestone, warn if already exists
        if atIndex==None:
            atIndex = len(self.funnels[funnel])
        for name in names:
            self.add_milestone(name, funnel, atIndex)            
            atIndex+=1
        return

    def add_milestone(self, name, funnel="default", atIndex=None):
        # add milestone, warn if already exists
        if name in self.milestones:
            print("warning: milestone '{}' is already defined".format(name))
        if atIndex==None:
            atIndex = len(self.funnels[funnel])
        self.funnels[funnel].insert(atIndex, name)
        self.milestones[name] = name
        return

    def add_node_to_milestone(self, dialog_node_id, milestone):
        # validate that node exists and milestone exists, warn if previous mapping exists
        if not dialog_node_id in self.dialog_node_index:
            print("Warning: dialog node '{}' does not exist in workspace".format(dialog_node_id))
        if not milestone in self.milestones:
            print("Error: milestone is not defined.  Use add_milestone() to add milestones")
            return
        if dialog_node_id in self.nodes:
            print("Warning: dialog node '{}' is already defined for milestone '{}', overriding...".format(dialog_node_id,self.nodes[dialog_node_id]))
        self.nodes[dialog_node_id] = milestone  

    def get_funnel(self, funnel="default"):
        if funnel not in self.funnels.keys():
            print("Error: funnel '{}' is not defined.".format(funnel))
            return
        return self.funnels[funnel]

    def get_milestone_in_nodes_visited(self, nodes):
        result = None
        for node in nodes:
            if node in self.nodes.keys():
                result = self.nodes[node]
                break
        return result      

    def enrich_milestones(self, df_logs):
        # take original data, for every nodes_visited, check if one has milestone def, enrich milestone
        # add column: milestone, funnels, 
        df_logs['milestone'] = None
        conversations_grouped =  df_logs.groupby("conversation_id")
        from tqdm import tqdm, tqdm_notebook
        if cat.use_widgets():
            slider = tqdm_notebook(range(100), desc="processing {} conversations".format(len(conversations_grouped)), ncols=900)
        else:
            slider = tqdm(range(100), desc="processing {} conversations".format(len(conversations_grouped)), ncols=900)
        i=0
        #for each conversation 
        for conversation, steps_df in conversations_grouped:
            i +=1
            if i/100 == int(i/100): # every 100 conversations update the slider
                step = int(100/len(conversations_grouped)*100)
                self._slider_update(slider, step, "Processed {} of {} conversations".format(i, len(conversations_grouped)))                  
            #nodes_visited_in_conversation = []
            steps_df= steps_df.sort_values(["response_timestamp"])
            last_index = None
            #for each conversation step, update the milestone in the original data
            for index, row in steps_df.iterrows():
                last_index = index
                nodes_visited = row["nodes_visited"]
                milestone = self.get_milestone_in_nodes_visited(nodes_visited)
                if milestone != None:
                    df_logs.at[index, "milestone"] = milestone   
            # if last step is not a milestone, add the Other node
            if df_logs.loc[last_index]['milestone'] == None:
                df_logs.at[index, "milestone"] = 'Other'
                #df_logs.loc[last_index]['milestone'] = 'Other'
        self._slider_update(slider, 100, "Processed {} conversations".format(i))
                          
    def _slider_update(self, slider, step, step_desc, step_message = None):
        width = 90
        prefix = ""
        long_message = prefix + step_desc + '.' * (width - len(step_desc) - len(prefix))
        slider.update(step)
        if step_message is not None:
            slider.write(step_message)
        slider.set_description(long_message)
        

def compute_flows(df, config):
    """
    DEPRECATED: compute aggregated flows across nodes and corresponding statistics, such as number of dropoffs
    returns results as a df for visualization
    """
    print("Error. compute_flows is deprecated.  Use aggregate_flows instead")
    return
    # check optional config attributes
    if "max_path_limit" not in config:
        print ("Warning, max_path_limit is missing from config.  Default 20 is assumed")
        config["max_path_limit"] = 20

    # check mandatory dataframe fields
    # TODO: check additional mandatory fields , eg log_id
    mandatory_columns = ['conversation_id', 'node_visited', 'response_timestamp']
    for column in mandatory_columns:
        if column not in df.columns.values:
            raise Exception("input data is missing mandatory column: " + column)

    analysis = _conversationPathAnalysis(config)
    for idx, conversation_df in df.groupby(df['conversation_id']):
        conversation_df = conversation_df.sort_values("response_timestamp")
        analysis.handle_conversation_path(conversation_df)

    df_node_out = pd.DataFrame.from_dict(analysis.nodes_paths, orient="index")
    df_node_out.reset_index(inplace=True)
    df_node_out.rename(columns={'index':'path'}, inplace=True)

    return pd.concat([df_node_out])[['path','name','type','is_conversation_start','flows','rerouted','dropped_off','conversation_log_ids_rerouted','conversation_log_ids_dropped_off','path_length']]

def aggregate_flows(df, max_depth=30, mode="turn-based", on_column="turn_label", trim_reroutes=False ):
    """
    compute aggregated flows across nodes and corresponding statistics, such as number of dropoffs
    returns results as a df for visualization
    """
    if mode not in ["turn-based", "milestone-based"]:
        print("invalid mode: {}.  Valid values are either turn-based or milestone-based".format(mode))
        return
    if on_column not in df.columns:
        print("invalid column name: {} does not exist in dataframe".format(column))
        return

    # check mandatory dataframe fields
    # TODO: check additional mandatory fields , eg log_id
    mandatory_columns = ['log_id', 'conversation_id', 'response_timestamp']
    for column in mandatory_columns:
        if column not in df.columns.values:
            raise Exception("input data is missing mandatory column: " + column)

    analysis = _conversationPathAnalysis(mode, on_column, max_depth, trim_reroutes)
    for idx, conversation_df in df.groupby(df['conversation_id']):
        conversation_df = conversation_df.sort_values("response_timestamp")
        analysis.handle_conversation_path(conversation_df)

    df_node_out = pd.DataFrame.from_dict(analysis.nodes_paths, orient="index")
    df_node_out.reset_index(inplace=True)
    df_node_out.rename(columns={'index':'path'}, inplace=True)

    return pd.concat([df_node_out])[['path','name','type','is_conversation_start','flows','rerouted','dropped_off','conversation_log_ids_rerouted','conversation_log_ids_dropped_off','path_length']]
 
def _find_consecutive_flow_states(df, column="milestone"):
    log_ids_to_delete = []
    for idx, conversation_df in df.groupby(df['conversation_id']):
        conversation_df = conversation_df.sort_values("response_timestamp")
        #for each conversation, remove duplicate milestones
        last_milestone = None
        for index, row in conversation_df.iterrows():
            if row[column] != last_milestone:
                last_milestone = row[column]
            else:
                log_ids_to_delete.append(row['log_id'])
    return log_ids_to_delete

def simplify_flow_consecutive_milestones(df):
    """
    remove consecutive milestones from the dataframe, to create a simplified flow visualization.
    """

    rows_to_delete = _find_consecutive_flow_states(df)
    print("Removed {} duplicate milestone rows".format(str(len(rows_to_delete))))
    result = df[~df["log_id"].isin(rows_to_delete)]
    return result


# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from IPython.display import display, HTML
import pandas as pd
from numpy import nan
import time

class ChainFilter(): 

    def __init__(self, df, description="<<ALL-DATA>>"):
        self.df = df
        self.description = ""
        self.filters = []
        self._append_filter(self.df, "<<ALL-DATA>>", 0)

    def _append_filter(self, df, filter_name, duration_sec):       
        users = conversations = logs = timestamp_from = timestamp_to = nan
        if "conversation_id" in df.columns:
            conversations = len(df["conversation_id"].unique())
        if "log_id" in df.columns:
            logs = len(df["log_id"].unique())
        if "user_id" in df.columns:
            users = len(df["user_id"].unique())
        if len(df) > 0:
            if "response_timestamp" in df.columns:
                timestamp_from = min(df["response_timestamp"])
            if "response_timestamp" in df.columns:
                timestamp_to = max(df["response_timestamp"])
        self.filters.append({"filter_name": filter_name, 
                             "filter_duration_sec": duration_sec,
                             "rows":len(df), 
                             "users": users, 
                             "conversations": conversations,
                             "timestamp_from": timestamp_from,
                             "timestamp_to": timestamp_to,
                             "df": df})
        self.df = df

    def setDescription(self, description):
        self.description = description
        return self

    def printConversationFilters(self):
        display(HTML("<h1>" + self.description + "</h1>"))
        display(HTML(pd.DataFrame(self.filters)[["filter_name", "filter_duration_sec", "rows", "users", "conversations", "timestamp_from","timestamp_to"]].to_html()))
        return self

    def getDataFrame(self, i=-1):
        return self.filters[i]["df"]   

    def getLineageDetails(self, i=-1):
        if i == -1:
            i = len(self.filters)-1
        result = ""
        for j in range(0,i+1):
            if j >0:
                result+= " --> "
            result+= self.filters[j]["filter_name"] + "(" + str(self.filters[j]["conversations"]) + ")"
        return result             
        
    def by_dialog_node_id(self, dialog_node_id):
        """
        filters conversations that include node_id in nodes_visited list
        """
        now = time.time()
        filtered_df = self.df
        #find which rows include the dialog node
        rows = filtered_df[filtered_df.apply(lambda x: dialog_node_id in x["nodes_visited"], axis=1)]
        #find their conversation_id
        unique_conversations = rows["conversation_id"].unique()  
        #filter by conversation
        filtered_df = filtered_df[filtered_df.apply(lambda x: x["conversation_id"] in unique_conversations, axis=1)]
        later = time.time()
        duration_sec = int(later - now)
        self._append_filter(filtered_df, "by_dialog_node_id (" + dialog_node_id + ")", duration_sec)
        return self

    def by_dialog_node_str(self, dialog_node_str):
        """
        filters conversations that include a node string (title/condition) value in the nodes_visited_str list
        """
        now = time.time()
        filtered_df = self.df
        #find which rows include the dialog node str
        rows = filtered_df[filtered_df.apply(lambda x: dialog_node_str in x["nodes_visited_str"], axis=1)]
        #find their conversation_id
        unique_conversations = rows["conversation_id"].unique()  
        #filter by conversation
        filtered_df = filtered_df[filtered_df.apply(lambda x: x["conversation_id"] in unique_conversations, axis=1)]
        later = time.time()
        duration_sec = int(later - now)
        self._append_filter(filtered_df, "by_dialog_node_str (" + dialog_node_str + ")", duration_sec)
        return self
    
    def by_turn_label(self, label):
        """
        return the rows of conversations that include a value in turn_label
        """
        now = time.time()
        filtered_df = self.df
        #find which rows include the dialog node
        #rows = filtered_df[filtered_df.apply(lambda x: turn_label in x["node_visited"], axis=1)]
        rows = filtered_df[filtered_df["turn_label"] == label]
        #find their conversation_id
        unique_conversations = rows["conversation_id"].unique()  
        #filter by conversation
        filtered_df = filtered_df[filtered_df.apply(lambda x: x["conversation_id"] in unique_conversations, axis=1)]
        later = time.time()
        duration_sec = int(later - now)
        self._append_filter(filtered_df, "by_turn_label (" + label + ")", duration_sec)
        return self

    def by_node_name(self, node_name, force_deprecation=False):
        """
        DEPRECATED: return the rows of conversations that include node in node_visited
        """
        if force_deprecation == False:
            print("Error: by_node_name() is deprecated and will be removed in the future. Use by_turn_label() or by_node_id() instead")
            print("For back compatibility, you can use the force_deprecation=True parameter, but make sure to have a 'node_visited' column in your data frame")
            return
        else:
             print("Warning: by_node_name() is deprecated and will be removed in the future. Use by_turn_label() or by_node_id() instead")
        now = time.time()
        filtered_df = self.df
        #find which rows include the dialog node
        rows = filtered_df[filtered_df.apply(lambda x: node_name in x["node_visited"], axis=1)]
        #find their conversation_id
        unique_conversations = rows["conversation_id"].unique()  
        #filter by conversation
        filtered_df = filtered_df[filtered_df.apply(lambda x: x["conversation_id"] in unique_conversations, axis=1)]
        later = time.time()
        duration_sec = int(later - now)
        self._append_filter(filtered_df, "by_node_name (" + node_name + ")", duration_sec)
        return self
    
    def trim_from_turn_label(self, turn_label):
        """
        filter and trim conversations steps prior to first step that includes a turn label
        """
        now = time.time()
        # create an empty dataframe with the same column names and types
        filtered_df = pd.DataFrame(data=None, columns=self.df.columns)
        for column in filtered_df.columns:
            filtered_df[column] = filtered_df[column].astype(self.df[column].dtypes.name)
        df_by_conversation_id = self.df.groupby(by="conversation_id")
        for conversation_id, conversation_df in df_by_conversation_id:
            i=0
            conversation_df = conversation_df.sort_values(by=["response_timestamp"])
            for index, row in conversation_df.iterrows():
                i=i+1
                if turn_label == row["turn_label"]:
                    num_of_elements_to_copy = len(conversation_df)-i+1
                    filtered_df = pd.concat([filtered_df,conversation_df.tail(num_of_elements_to_copy)])
                    break
        later = time.time()
        duration_sec = int(later - now)
        self._append_filter(filtered_df, "trim_from_turn_label (" + turn_label + ")", duration_sec)
        return self

    def trim_from_node_id(self, node_id):
        """
        filter and trim conversations steps prior to first step that includes node_id in nodes_visited.
        """
        now = time.time()
     
        # create an empty dataframe with the same column names and types
        filtered_df = pd.DataFrame(data=None, columns=self.df.columns)
        for column in filtered_df.columns:
            filtered_df[column] = filtered_df[column].astype(self.df[column].dtypes.name)
        df_by_conversation_id = self.df.groupby(by="conversation_id")
        for conversation_id, conversation_df in df_by_conversation_id:
            i=0
            conversation_df = conversation_df.sort_values(by=["response_timestamp"])
            for index, row in conversation_df.iterrows():
                i=i+1
                nodes_visited = row["nodes_visited"]
                if node_id in nodes_visited:
                    num_of_elements_to_copy = len(conversation_df)-i+1
                    filtered_df = pd.concat([filtered_df,conversation_df.tail(num_of_elements_to_copy)])
                    break
        later = time.time()
        duration_sec = int(later - now)
        self._append_filter(filtered_df, "trim_from_node_id (" + node_id + ")", duration_sec)
        return self

    def from_node_onwards(self, node_name):
        """
        DEPCRECATED: filter & trim conversations starting from node_id (in node_visited) 
        """
        print("Error: from_node_onwards() is deprecated. Use trim_from_turn_label() or trim_from_node_id() instead")
        return
        now = time.time()
     
        # create an empty dataframe with the same column names and types
        filtered_df = pd.DataFrame(data=None, columns=self.df.columns)
        for column in filtered_df.columns:
            filtered_df[column] = filtered_df[column].astype(self.df[column].dtypes.name)
        df_by_conversation_id = self.df.groupby(by="conversation_id")
        #df_by_conversation_id = filtered_df.sort_values(by=["response_timestamp"]).groupby(by="conversation_id")
        for conversation_id, conversation_df in df_by_conversation_id:
            i=0
            conversation_df = conversation_df.sort_values(by=["response_timestamp"])
            for index, row in conversation_df.iterrows():
                i=i+1
                node_visited = row["node_visited"]
                if node_name == node_visited:
                    num_of_elements_to_copy = len(conversation_df)-i+1
                    filtered_df = pd.concat([filtered_df,conversation_df.tail(num_of_elements_to_copy)])
                    break
        later = time.time()
        duration_sec = int(later - now)
        self._append_filter(filtered_df, "from_node_onwards (" + node_name + ")", duration_sec)
        return self
    
    def by_date_range(self, start_date, end_date):
        now = time.time()        
        mask = (self.df['response_timestamp'] >= start_date) & (self.df['response_timestamp'] <= end_date)
        filtered_df = self.df.loc[mask]
        later = time.time()
        duration_sec = int(later - now)
        self._append_filter(filtered_df, "by_date_range (" + str(start_date) + ", " + str(end_date) + ")", duration_sec)
        return self



# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np
import datetime

def by_included_node(node, df_logs_formated):
    """
    return subset of conversations that contains specific node
    """
    print("filtering.by_included_node() is DEPRECATED and will be removed in a future release.  Use filtering2.by_node_id() instead ")
    
    # create an empty dataframe with the same column names and types
    filtered_df_logs = pd.DataFrame(data=None, columns=df_logs_formated.columns)
    for column in filtered_df_logs.columns:
        filtered_df_logs[column] = filtered_df_logs[column].astype(df_logs_formated[column].dtypes.name)

    # TODO: should we sort?
    df_by_conversation_id = df_logs_formated.sort_values(by=["response_timestamp"]).groupby(by="conversation_id")
    # TODO: potentially iterate over groupby
    ConvDict = {}
    for key, group in df_by_conversation_id:
        ConvDict[key] = group

    for i, (k, v) in enumerate(ConvDict.items()):
        # consider if we need to sort
        dfP1=pd.DataFrame(v, columns=v.columns).sort_values(by=['response_timestamp'])
        if any(dfP1['node_visited'] == node):
            filtered_df_logs = pd.concat([filtered_df_logs,dfP1])

    if filtered_df_logs.empty:
        print('Filtering yielded an empty dataframe. Path flow analysis requires a non-empty dataframe.')

    print('Initial amount of records (before filtering):', len(df_logs_formated))
    print('Amount of filtered records:', len(df_logs_formated)-len(filtered_df_logs))
    print('Final amount of records (after filtering):', len(filtered_df_logs))

    return filtered_df_logs


def by_included_absolute_path(path, df_logs_formated):
    """
    return subset of conversations that contains specific node
    @param path path: array
    """
    print("filtering.by_included_absolute_path() is DEPRECATED and will be removed in a future release. ")
    # create an empty dataframe with the same column names and types
    filtered_df_logs = pd.DataFrame(data=None, columns=df_logs_formated.columns)
    for column in filtered_df_logs.columns:
        filtered_df_logs[column] = filtered_df_logs[column].astype(df_logs_formated[column].dtypes.name)

    # TODO: should we sort?
    df_by_conversation_id = df_logs_formated.sort_values(by=["response_timestamp"]).groupby(by="conversation_id")
    # TODO: potentially iterate over groupby
    ConvDict = {}
    for key, group in df_by_conversation_id:
        ConvDict[key] = group

    for i, (k, v) in enumerate(ConvDict.items()):
        # consider if we need to sort
        dfP1=pd.DataFrame(v, columns=v.columns).sort_values(by=['response_timestamp'])
        # length of conversation is at least length of path
        if len(dfP1) >= len(path):
            for i in range(len(path)):
                if path[i] != dfP1.iloc[i]['node_visited']:
                    break
                # last element
                if i == len(path)-1:
                     filtered_df_logs = pd.concat([filtered_df_logs,dfP1])

    if filtered_df_logs.empty:
        print('Filtering yielded an empty dataframe. Path flow analysis requires a non-empty dataframe.')

    print('Initial amount of records (before filtering):', len(df_logs_formated))
    print('Amount of filtered records:', len(df_logs_formated)-len(filtered_df_logs))
    print('Final amount of records (after filtering):', len(filtered_df_logs))

    return filtered_df_logs

def by_initial_intent(intent, df_logs_formated):
    """
    filter conversations starting with initial intent
    """
    print("filtering.by_initial_intent() is DEPRECATED and will be removed in a future release. Use filtering2.by_node_id() instead.")
    
    # create an empty dataframe with the same column names and types
    filtered_df_logs = pd.DataFrame(data=None, columns=df_logs_formated.columns)
    for column in filtered_df_logs.columns:
        filtered_df_logs[column] = filtered_df_logs[column].astype(df_logs_formated[column].dtypes.name)
    df_by_conversation_id = df_logs_formated.sort_values(by=["response_timestamp"]).groupby(by="conversation_id")

    ConvDict = {}
    for key, group in df_by_conversation_id:
        ConvDict[key] = group

    for i, (k, v) in enumerate(ConvDict.items()):
        dfP1=pd.DataFrame(v, columns=v.columns).sort_values(by=['response_timestamp'])
        if (dfP1['intent'].iloc[0] == intent):
            filtered_df_logs = pd.concat([filtered_df_logs,dfP1])
    if filtered_df_logs.empty:
        print('Filtering yielded an empty dataframe. Path flow analysis requires a non-empty dataframe.')

    print('Initial amount of records (before filtering):', len(df_logs_formated))
    print('Amount of filtered records:', len(df_logs_formated)-len(filtered_df_logs))
    print('Final amount of records (after filtering):', len(filtered_df_logs))

    return filtered_df_logs

def from_node_onwards(node, df_logs_formated):
    """
    filtering by truncating conversations from selected node onwards
    """
    print("filtering.from_node_onwards is DEPRECATED and will be removed in a future release. Use filtering2.trim_from_node_id instead.")
    # create an empty dataframe with the same column names and types
    filtered_df_logs = pd.DataFrame(data=None, columns=df_logs_formated.columns)
    for column in filtered_df_logs.columns:
        filtered_df_logs[column] = filtered_df_logs[column].astype(df_logs_formated[column].dtypes.name)

    ## TODO: should we move sorting into the loop? 
    df_by_conversation_id = df_logs_formated.sort_values(by=["response_timestamp"]).groupby(by="conversation_id")
    for conversation_id, conversation_df in df_by_conversation_id:
        i=0
        for index, row in conversation_df.iterrows():
            i=i+1
            node_visited = row["node_visited"]

            if node == node_visited:
                num_of_elements_to_copy = len(conversation_df)-i+1
                filtered_df_logs = pd.concat([filtered_df_logs,conversation_df.tail(num_of_elements_to_copy)])
                break
    if filtered_df_logs.empty:
        print('Filtering yielded an empty dataframe. Path flow analysis requires a non-empty dataframe.')

    print('Initial amount of records (before filtering):', len(df_logs_formated))
    print('Amount of filtered records:', len(df_logs_formated)-len(filtered_df_logs))
    print('Final amount of records (after filtering):', len(filtered_df_logs))

    return filtered_df_logs

def by_date_range(df, start_date, end_date):
    print("filtering.by_date_range is DEPRECATED and will be removed in a future release. Use filtering2.by_date_range instead.")

    mask = (df['response_timestamp'] >= start_date) & (df['response_timestamp'] <= end_date)
    df_1 = df.loc[mask].reset_index()

    if df_1.empty:
        print('Filtering yielded an empty dataframe. Path flow analysis requires a non-empty dataframe.')

    print('Initial amount of records (before filtering):', len(df))
    print('Amount of filtered records:', len(df)-len(df_1))
    print('Final amount of records (after filtering):', len(df_1))

    return df_1


# (C) Copyright IBM Corp. 2019, 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class __Config__:
  def __init__(self):
    self.globals = {
        'enable_widgets': False,
        'forced_production_mode': False
    }
  def set(self, key, value):
      self.globals[key] = value
  def get(self, key):
      return self.globals[key]

__config__ = __Config__()


def disable_widgets():
    """
    Disable interactive UI using ipywidgets (require nb_extenstions)
    """
    __config__.set('enable_widgets', False)
    print("interactive UI widgets will be disabled")


def enable_widgets():
    """
    Enable interactive UI using ipywidgets (require nb_extenstions)
    """
    __config__.set('enable_widgets', True)
    print("interactive UI widgets will be enabled")

def use_widgets():
    return __config__.get('enable_widgets')

def force_production_mode():
    """
    Production mode should be autodetected, but if not, as a workaround, you can force it
    """
    __config__.set('forced_production_mode', True)
    print("Production mode will be forced (instead of autodetected)")

def is_forced_production_mode():
    return __config__.get('forced_production_mode')

def get_version():
    from .__meta__ import __version__
    return __version__

def is_production():
    import subprocess
    import sys

    reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
    installed_packages = [r.decode().split('==')[0] for r in reqs.split()]
    return 'conversation-analytics-toolkit' in installed_packages

def _get_src_project_folder():
    import os, inspect
    current_src_location = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    project_folder = current_src_location[:(current_src_location.rfind("src")-1)]
    return project_folder

if not is_production() and not is_forced_production_mode():
    project_folder = _get_src_project_folder()
    print('dev-mode detected:' + project_folder)
else:
    print('production-mode detected')

print('version: ' + get_version())

__version__ = get_version()
